# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Timing support
"""

from __future__ import annotations

import time
from contextlib import contextmanager

from .formats import friendly_duration


class Stopwatch:
    """
    Timer class for recording elapsed wall time in operations.
    """

    acc_time: float | None = None
    start_time: float | None = None
    stop_time: float | None = None

    def __init__(self, start=True):
        if start:
            self.start()

    def start(self):
        self.start_time = time.perf_counter()

    def stop(self):
        self.stop_time = time.perf_counter()
        if self.start_time is not None and self.acc_time is not None:
            self.acc_time += self.stop_time - self.start_time

    def elapsed(self, *, accumulated=True) -> float:
        """
        Get the elapsed time.
        """
        assert self.start_time is not None
        stop = self.stop_time or time.perf_counter()

        t = stop - self.start_time
        if accumulated and self.acc_time is not None:
            t += self.acc_time

        return t

    @contextmanager
    def measure(self, accumulate: bool = False):
        """
        Context manager to measure an item, optionally accumulating its time.
        """
        if accumulate:
            if self.acc_time is None:
                self.acc_time = 0.0
        else:
            self.acc_time = None

        self.start()
        try:
            yield self
        finally:
            self.stop()

    def __str__(self):
        elapsed = self.elapsed()
        return friendly_duration(elapsed)

    def __repr__(self):
        elapsed = self.elapsed()
        if self.stop_time:
            return "<Stopwatch stopped at {:.3f}s>".format(elapsed)
        else:
            return "<Stopwatch running at {:.3f}s>".format(elapsed)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import numpy as np

import hypothesis.extra.numpy as nph
import hypothesis.strategies as st
from hypothesis import given, settings

from lenskit.stats import argtopn


def test_simple_topn():
    positions = argtopn([1.0, 0.0], 1)
    assert len(positions) == 1
    assert positions[0] == 0


def test_simple_topn_rev():
    positions = argtopn([0.0, 1.0], 1)
    assert len(positions) == 1
    assert positions[0] == 1


@given(
    nph.arrays(nph.floating_dtypes(endianness="="), st.integers(0, 5000)), st.integers(min_value=-1)
)
def test_arg_topn(xs, k):
    positions = argtopn(xs, k)
    if k >= 0:
        assert len(positions) <= k
    assert positions.dtype == np.int64
    if k == 0 or np.all(np.isnan(xs)):
        assert len(positions) == 0
        return

    top_xs = xs[positions]

    sort_xs = np.sort(-xs[~np.isnan(xs)])

    # we have the correct number of positions
    if k >= 0:
        assert len(positions) == min(k, np.sum(~np.isnan(xs)))
    else:
        assert len(positions) == np.sum(~np.isnan(xs))
    # all rank positions are valid
    assert np.all(positions >= 0)
    assert np.all(positions < len(xs))
    # all rank positions are unique
    assert len(np.unique(positions)) == len(positions)
    # all ranked items are numbers
    assert not np.any(np.isnan(top_xs))

    # we have the largest values
    if len(positions) < k:
        omitted = np.ones(len(xs), dtype="bool")
        omitted[positions] = False
        if not np.all(np.isnan(xs[omitted])):
            assert np.all(top_xs >= np.nanmax(xs[omitted]))

    # the values are sorted
    if len(top_xs) > 1:
        assert np.all(top_xs[:-1] >= top_xs[1:])

    # the min matches the underlying sort
    if k >= 1:
        assert top_xs[-1] == -sort_xs[min(k - 1, len(sort_xs) - 1)]
    elif np.all(np.isfinite(sort_xs)):
        assert np.all(top_xs == -sort_xs)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import numpy as np
import pyarrow as pa
import torch
from numpy.typing import NDArray

import hypothesis.extra.numpy as nph
import hypothesis.strategies as st
from hypothesis import assume, given

from lenskit.data.mtarray import MTArray, MTIntArray


@given(
    nph.arrays(
        dtype=st.one_of(nph.integer_dtypes(endianness="="), nph.floating_dtypes(endianness="=")),
        shape=nph.array_shapes(),
    )
)
def test_from_numpy(arr: NDArray[np.generic]):
    # limit to data types that match
    assume(np.all(np.isfinite(arr)))
    mta = MTArray(arr)
    assert mta.shape == arr.shape
    npa = mta.numpy()
    assert npa is arr

    tensor = mta.torch()
    assert tensor.shape == arr.shape
    assert np.all(tensor.numpy() == arr)

    assert np.asarray(mta) is arr


@given(
    nph.arrays(
        dtype=st.one_of(nph.integer_dtypes(endianness="="), nph.floating_dtypes(endianness="=")),
        shape=nph.array_shapes(),
    )
)
def test_from_torch(arr: NDArray[np.generic]):
    # limit to data types that match
    assume(np.all(np.isfinite(arr)))
    ot = torch.from_numpy(arr)
    mta = MTArray(ot)
    assert mta.shape == arr.shape
    tensor = mta.torch()
    assert tensor is ot

    npa = mta.numpy()
    assert npa.shape == arr.shape
    assert np.all(npa == arr)


@given(
    nph.arrays(
        dtype=st.one_of(nph.integer_dtypes(endianness="="), nph.floating_dtypes(endianness="=")),
        shape=nph.array_shapes(),
    )
)
def test_from_arrow(arr: NDArray[np.generic]):
    # limit to data types that match
    assume(np.all(np.isfinite(arr)))
    if len(arr.shape) == 1:
        oarr = pa.array(arr)
    else:
        oarr = pa.Tensor.from_numpy(arr)
    mta = MTArray(oarr)
    assert mta.shape == arr.shape

    a2 = mta.arrow()
    assert a2 is oarr

    npa = mta.numpy()
    assert npa.shape == arr.shape
    assert np.all(npa == arr)


@given(
    nph.arrays(
        dtype=st.one_of(nph.integer_dtypes(endianness="="), nph.floating_dtypes(endianness="=")),
        shape=nph.array_shapes(),
    ),
    st.sampled_from(["numpy", "arrow", "torch"]),
    st.sampled_from(["numpy", "arrow", "torch"]),
)
def test_data_combo(arr: NDArray[np.generic], src, tgt):
    # limit to data types that match
    assume(np.all(np.isfinite(arr)))

    match src:
        case "numpy":
            mta = MTArray(arr)
        case "torch":
            mta = MTArray(torch.from_numpy(arr))
        case "arrow" if len(arr.shape) == 1:
            a2 = pa.array(arr)
            assert isinstance(a2, pa.Array)
            mta = MTArray(a2)
        case "arrow":
            mta = MTArray(pa.Tensor.from_numpy(arr))

    assert not isinstance(mta._convertible(), pa.Scalar)

    assert mta.shape == arr.shape

    try:
        match tgt:
            case "numpy":
                out = mta.numpy()
            case "arrow":
                out = mta.arrow()
            case "torch":
                out = mta.torch()
    except Exception as e:
        print("internal:", type(mta._unknown), mta._unknown)
        raise e

    if isinstance(out, pa.Array):
        assert len(arr.shape) == 1
        assert len(out) == arr.shape[0]
    else:
        assert out.shape == arr.shape
    assert np.all(np.asarray(out) == arr)


@given(st.lists(st.integers(min_value=np.iinfo(np.int64).min, max_value=np.iinfo(np.int64).max)))
def test_from_list(xs: list[int]):
    # limit to data types that match
    mta = MTIntArray(xs)
    assert mta.shape == (len(xs),)

    npa = mta.numpy()
    assert np.all(npa == xs)

    tensor = mta.torch()
    assert np.all(tensor.numpy() == xs)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Whole-pipeline integration tests to check for composite bugs.
"""

import numpy as np

from pytest import fixture

from lenskit import batch, recommend
from lenskit.als import ImplicitMFScorer
from lenskit.basic import (
    RandomSelector,
    UnratedTrainingItemsCandidateSelector,
    UserTrainingHistoryLookup,
)
from lenskit.data import Dataset, ItemList, QueryInput, RecQuery
from lenskit.pipeline import Pipeline, PipelineBuilder, topn_pipeline


@fixture(scope="module")
def random_pipe(ml_ds: Dataset):
    pb = PipelineBuilder()
    query = pb.create_input("query", QueryInput)
    history = pb.add_component("history-lookup", UserTrainingHistoryLookup, query=query)
    candidates = pb.add_component(
        "candidate-selector", UnratedTrainingItemsCandidateSelector, query=history
    )
    pb.add_component("recommender", RandomSelector, {"n": 100}, items=candidates)

    pipe = pb.build()
    pipe.train(ml_ds)
    yield pipe


def test_history_correct(rng: np.random.Generator, ml_ds: Dataset, random_pipe: Pipeline):
    users = rng.choice(ml_ds.users.ids(), 200, replace=False)

    for uid in users:
        train_items = ml_ds.user_row(uid)
        assert train_items is not None

        query = random_pipe.run("history-lookup", query=uid)
        assert isinstance(query, RecQuery)
        assert query.user_id == uid
        user_items = query.user_items
        assert isinstance(user_items, ItemList)
        assert len(user_items) == len(train_items)
        assert set(user_items.ids()) == set(train_items.ids())
        assert len(user_items) == len(set(user_items.ids()))


def test_candidates_correct(rng: np.random.Generator, ml_ds: Dataset, random_pipe: Pipeline):
    users = rng.choice(ml_ds.users.ids(), 200, replace=False)

    for uid in users:
        train_items = ml_ds.user_row(uid)
        assert train_items is not None

        candidates = random_pipe.run("candidate-selector", query=uid)
        assert isinstance(candidates, ItemList)
        # make sure the candidates don't have any training items
        assert len(candidates) == ml_ds.item_count - len(train_items)
        assert len(candidates) == len(set(candidates.ids()))


def test_training_items_removed_solo(
    rng: np.random.Generator, ml_ds: Dataset, random_pipe: Pipeline
):
    users = rng.choice(ml_ds.users.ids(), 200, replace=False)

    for uid in users:
        items = recommend(random_pipe, uid)

        train_items = ml_ds.user_row(uid)
        assert train_items is not None

        assert len(items) == 100
        assert not np.any(np.isin(items.ids(), train_items.ids()))
        assert not np.any(np.isin(items.numbers(), train_items.numbers()))
        assert len(items) == len(set(items.ids()))


def test_training_items_removed_batch(
    rng: np.random.Generator, ml_ds: Dataset, random_pipe: Pipeline
):
    users = rng.choice(ml_ds.users.ids(), 200, replace=False)
    recs = batch.recommend(random_pipe, users, n_jobs=1)

    for key, items in recs.items():
        uid = key.user_id
        assert uid in users
        train_items = ml_ds.user_row(uid)
        assert train_items is not None

        assert len(items) == 100
        # make sure we didn't recommend any training items
        assert not np.any(np.isin(items.ids(), train_items.ids()))
        assert not np.any(np.isin(items.numbers(), train_items.numbers()))
        assert len(items) == len(set(items.ids()))


def test_training_items_removed_scorer(rng: np.random.Generator, ml_ds: Dataset):
    pipe = topn_pipeline(scorer=ImplicitMFScorer, n=100)
    pipe.train(ml_ds)

    users = rng.choice(ml_ds.users.ids(), 200, replace=False)
    recs = batch.recommend(pipe, users, n_jobs=1)

    for key, items in recs.items():
        uid = key.user_id
        assert uid in users
        train_items = ml_ds.user_row(uid)
        assert train_items is not None

        assert len(items) == 100
        # make sure we didn't recommend any training items
        assert not np.any(np.isin(items.ids(), train_items.ids()))
        assert not np.any(np.isin(items.numbers(), train_items.numbers()))
        assert len(items) == len(set(items.ids()))



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from __future__ import annotations

from dataclasses import dataclass
from typing import Generic, Literal, TypeAlias, TypeVar

import pandas as pd

from lenskit.data import Dataset, ItemListCollection

SplitTable: TypeAlias = Literal["matrix"]
TK = TypeVar("TK", bound=tuple)


@dataclass
class TTSplit(Generic[TK]):
    """
    A train-test set from splitting or other sources.

    Stability:
        Caller
    """

    train: Dataset
    """
    The training data.
    """

    test: ItemListCollection[TK]
    """
    The test data.
    """

    name: str | None = None
    """
    A name for this train-test split.
    """

    @property
    def test_size(self) -> int:
        """
        Get the number of test pairs.
        """
        return sum(len(il) for il in self.test.lists())

    @property
    def test_df(self) -> pd.DataFrame:
        """
        Get the test data as a data frame.
        """
        return self.test.to_df()

    @property
    def train_df(self) -> pd.DataFrame:
        """
        Get the training data as a data frame.
        """
        return self.train.interaction_matrix(format="pandas", field="all")



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from __future__ import annotations

import faulthandler
import multiprocessing as mp
import os
import warnings
from dataclasses import dataclass
from typing import Any

from typing_extensions import Generic

from lenskit.logging import get_logger

from .invoker import A, InvokeOp, M, R
from .serialize import SHMData, shm_deserialize

_log = get_logger(__name__)

__work_context: WorkerData


@dataclass(frozen=True)
class WorkerData(Generic[M, A, R]):
    func: InvokeOp[M, A, R]
    model: M


def initalize(ctx: SHMData) -> None:
    global __work_context, __progress
    faulthandler.enable()
    proc = mp.current_process()
    log = _log.bind(pid=proc.pid, pname=proc.name)

    warnings.filterwarnings("ignore", "Sparse CSR tensor support is in beta state", UserWarning)

    try:
        __work_context = shm_deserialize(ctx)
    except Exception as e:
        log.error("deserialization failed: %s", e)
        raise e

    log.debug("worker %d ready (process %s)", os.getpid(), mp.current_process().name)


def worker(arg: Any) -> Any:
    res = __work_context.func(__work_context.model, arg)
    return res



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import numpy as np

from hypothesis import assume, given
from hypothesis import strategies as st
from hypothesis.extra import numpy as nph
from pytest import approx, mark

from lenskit.logging import get_logger
from lenskit.stats import gini

_log = get_logger(__name__)


def test_gini_uniform():
    ones = np.ones(100)
    assert gini(ones) == approx(0)


def test_completely_unequal():
    arr = np.zeros(10000)
    arr[0] = 1000
    assert gini(arr) == approx(1, abs=0.001)


@mark.skipif(np.version.version < "2.0", reason="NumPy too old")
@given(
    nph.arrays(
        st.one_of(nph.floating_dtypes(sizes=[32, 64]), nph.integer_dtypes()),
        nph.array_shapes(max_dims=1, min_side=2),
        elements={
            "allow_nan": False,
            "allow_infinity": False,
            "min_value": 0,
            "allow_subnormal": False,
        },
    )
)
def test_random_values(xs):
    # max value here, large max_values don't work on elements
    assume(np.all(xs < 1e12))
    assume(np.any(xs > 0))

    g = gini(xs)
    assert g >= 0
    assert g <= 1

    # The gini should be equal to the difference between ideal area and actual
    # area under the curve.  We can approximate this with NumPy trapezoid.
    n = len(xs)
    ideal = np.sum(xs) * len(xs) * 0.5
    xss = np.zeros(n + 1, dtype="f8")
    xss[1:] = np.cumsum(np.sort(xs))
    actual = np.trapezoid(xss)

    _log.debug("computed gini", n=len(xs), gini=g, actual=actual, ideal=ideal)
    # we use max just to deal with extremely small values
    assert g == approx(max((ideal - actual) / ideal, 0), abs=0.001)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import numpy as np
import pandas as pd

import hypothesis.strategies as st
from hypothesis import given
from pytest import approx, mark

from lenskit.data import ItemList
from lenskit.metrics import call_metric
from lenskit.metrics.ranking import NDCG
from lenskit.metrics.ranking._dcg import array_dcg, fixed_dcg
from lenskit.testing import integer_ids


def test_ndcg_empty():
    recs = ItemList(ordered=True)
    truth = ItemList([1, 2, 3], rating=[3.0, 5.0, 4.0])
    assert call_metric(NDCG, recs, truth) == approx(0.0)


def test_ndcg_no_match():
    recs = ItemList([4], ordered=True)
    truth = ItemList([1, 2, 3], rating=[3.0, 5.0, 4.0])
    assert call_metric(NDCG, recs, truth) == approx(0.0)


def test_ndcg_perfect():
    recs = ItemList([2, 3, 1], ordered=True)
    truth = ItemList([1, 2, 3], rating=[3.0, 5.0, 4.0])
    assert call_metric(NDCG, recs, truth) == approx(1.0)


def test_ndcg_perfect_k_short():
    recs = ItemList([2, 3, 1], ordered=True)
    truth = ItemList([1, 2, 3], rating=[3.0, 5.0, 4.0])
    assert call_metric(NDCG, recs, truth, k=2) == approx(1.0)
    assert call_metric(NDCG, recs[:2], truth, k=2) == approx(1.0)


def test_ndcg_shorter_not_best():
    recs = ItemList([1, 2], ordered=True)
    truth = ItemList([1, 2, 3], rating=[3.0, 5.0, 4.0])
    b_ideal = fixed_dcg(3)
    r_ideal = array_dcg(np.array([5.0, 4.0, 3.0]))
    assert call_metric(NDCG, recs, truth) == approx(fixed_dcg(2) / b_ideal)
    assert call_metric(NDCG, recs, truth, k=2) == approx(1.0)
    assert call_metric(NDCG, recs, truth, gain="rating") == approx(
        array_dcg(np.array([3.0, 5.0])) / r_ideal
    )


def test_ndcg_perfect_k():
    recs = ItemList([2, 3], ordered=True)
    truth = ItemList([1, 2, 3], rating=[3.0, 5.0, 4.0])
    assert call_metric(NDCG, recs, truth, k=2) == approx(1.0)


def test_ndcg_perfect_k_norate():
    recs = ItemList([1, 3], ordered=True)
    truth = ItemList([1, 2, 3], rating=[3.0, 5.0, 4.0])
    assert call_metric(NDCG, recs, truth, k=2) == approx(1.0)


def test_ndcg_almost_perfect_k_gain():
    recs = ItemList([1, 3], ordered=True)
    truth = ItemList([1, 2, 3], rating=[3.0, 5.0, 4.0])
    assert call_metric(NDCG, recs, truth, k=2, gain="rating") == approx(
        array_dcg(np.array([3.0, 4.0])) / array_dcg(np.array([5.0, 4.0]))
    )


@given(
    st.lists(integer_ids(), min_size=1, max_size=100, unique=True),
    st.integers(1, 100),
)
def test_ndcg_alt_discount(items, k):
    rng = np.random.default_rng()
    picked = rng.choice(items, size=max(len(items) // 2, 1), replace=False)
    recs = ItemList(items, ordered=True)
    truth = ItemList(picked)

    mv_weighted = call_metric(NDCG, recs, truth, k=k)
    mv_legacy = call_metric(NDCG, recs, truth, k=k, discount=np.log2)

    try:
        assert mv_weighted == approx(mv_legacy)
    except Exception as e:
        e.add_note(f"recs: {recs}")
        e.add_note(f"truth: {truth}")
        raise e



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Support for logging from worker processes.

This is internal support code, clients and LensKit implementers can usually
ignore it.
"""

from __future__ import annotations

import copy
import logging
import multiprocessing as mp
import pickle
import warnings
from dataclasses import dataclass
from logging import Handler, LogRecord, getLogger
from threading import Lock
from time import perf_counter
from typing import Any, Self, overload
from uuid import UUID

import structlog
import zmq
from structlog.typing import EventDict

from lenskit.logging.progress._base import Progress

from .._limit import RateLimit
from .._proxy import get_logger
from ..config import CORE_PROCESSORS, active_logging_config, log_warning
from ..processors import add_process_info
from ..progress import set_progress_impl

# from ..progress._worker import ProgressMessage
from ..tasks import Task
from ..tracing import lenskit_filtering_logger
from ._protocol import (
    LogChannel,
    MsgAuthenticator,
    ProgressField,
    ProgressMessage,
)

_active_context: WorkerContext | None = None
_log = get_logger(__name__)


@dataclass
class WorkerLogConfig:
    """
    Configuration for worker logging.
    """

    address: str
    level: int
    authkey: bytes | None = None

    @classmethod
    @overload
    def current(cls) -> Self: ...
    @classmethod
    @overload
    def current(cls, *, from_monitor: bool = True) -> Self | None: ...
    @classmethod
    def current(cls, *, from_monitor: bool = True):
        """
        Get the current worker logging configuration.
        """

        if _active_context is not None:
            return _active_context.config
        elif from_monitor:
            from ..monitor import get_monitor

            mon = get_monitor()
            if mon.log_address is None:
                raise RuntimeError("monitor has no log address")
            cfg = active_logging_config()
            level = cfg.effective_level if cfg is not None else logging.INFO
            return cls(
                address=mon.log_address, level=level, authkey=bytes(mp.current_process().authkey)
            )
        else:
            return None


class WorkerContext:
    """
    Activate (and deactivate) a worker context.  This handles setup and teardown
    of logging, etc.

    Only one worker context can be active, regardless of how many threads are active.

    Stability:
        internal
    """

    config: WorkerLogConfig
    zmq: zmq.Context[zmq.Socket[bytes]]
    _log_handler: ZMQLogHandler
    _ref_count: int = 0
    _is_driver: bool = False

    def __init__(
        self,
        config: WorkerLogConfig,
        *,
        driver: bool = False,
        zmq: zmq.Context[zmq.Socket[bytes]] | None = None,
    ):
        self.config = config
        if self.config.authkey is None:
            self.config.authkey = mp.current_process().authkey
        self._is_driver = driver
        if zmq is not None:
            self.zmq = zmq

    @staticmethod
    def active() -> WorkerContext | None:
        return _active_context

    def start(self):
        """
        Start the logging context.
        """
        global _active_context
        if _active_context is not None:
            raise RuntimeError("worker context already active")
        _active_context = self

        if not self._is_driver:
            self.zmq = zmq.Context()

        self._log_handler = ZMQLogHandler(self.zmq, self.config)

        if self._is_driver:
            return

        root = getLogger()
        root.addHandler(self._log_handler)
        root.setLevel(self.config.level)

        structlog.configure(
            [add_process_info]
            + CORE_PROCESSORS
            + [structlog.processors.ExceptionPrettyPrinter(), self._log_handler.send_structlog],
            wrapper_class=lenskit_filtering_logger(self.config.level),
            logger_factory=structlog.stdlib.LoggerFactory(),
            cache_logger_on_first_use=False,
        )
        warnings.showwarning = log_warning
        set_progress_impl(WorkerProgress)
        _log.debug("log context activated")

    def shutdown(self):
        global _active_context
        if not self._is_driver:
            root = getLogger()
            root.removeHandler(self._log_handler)
            set_progress_impl(None)

        self._log_handler.shutdown()
        if not self._is_driver:
            self.zmq.term()

        _active_context = None

    def send_task(self, task: Task):
        self._log_handler.send_task(task)

    def send_progress(self, update: ProgressMessage):
        """
        Send a progress update event.
        """
        self._log_handler.send_progress(update)

    def send_record(self, sink_id: UUID, record: Any):
        """
        Send a record to a record sink.
        """
        self._log_handler.send_record(sink_id, record)

    def __enter__(self):
        if self._ref_count == 0:
            self.start()
        self._ref_count += 1
        return self

    def __exit__(self, *args):
        self._ref_count -= 1
        if self._ref_count == 0:
            self.shutdown()


class WorkerProgress(Progress):  # pragma: nocover
    """
    Progress logging over the pipe to a supervisor.
    """

    label: str
    context: WorkerContext | None
    completed: float = 0
    _fields: dict[str, str | None] = {}
    _limit: RateLimit

    def __init__(
        self,
        label: str,
        total: int | None,
        fields: dict[str, str | None],
    ):
        super().__init__()
        self.context = WorkerContext.active()
        self.label = label
        self.total = total
        self._fields = fields
        self._limit = RateLimit(20)

    def update(
        self,
        advance: int = 1,
        completed: int | None = None,
        total: int | None = None,
        **kwargs: float | int | str,
    ):
        if self.context is None:
            return

        if completed is not None:
            self.completed = completed
        else:
            self.completed += advance
        if total is not None:
            self.total = total

        now = perf_counter()
        if self._limit.want_update(now) or self.completed == self.total:
            fields = {
                name: ProgressField(value, self._fields[name])
                for (name, value) in kwargs.items()
                if name in kwargs
            }
            self.context.send_progress(
                ProgressMessage(
                    progress_id=self.uuid,
                    label=self.label,
                    total=self.total,
                    completed=self.completed,
                    fields=fields,
                )
            )
            self._limit.mark_update(now)

    def finish(self):
        if self.context is None:
            return

        self.context.send_progress(
            ProgressMessage(
                progress_id=self.uuid,
                label=self.label,
                total=self.total,
                completed=self.completed,
                finished=True,
            )
        )


class ZMQLogHandler(Handler):
    _lock: Lock
    socket: zmq.Socket[bytes]
    key: bytes
    _render = structlog.processors.JSONRenderer()
    _auth: MsgAuthenticator

    def __init__(self, zmq_context: zmq.Context, config: WorkerLogConfig):
        super().__init__()
        self.config = config
        self._lock = Lock()
        assert config.authkey is not None
        self._auth = MsgAuthenticator(config.authkey)
        self.socket = zmq_context.socket(zmq.PUSH)
        self.socket.connect(config.address)

    def handle(self, record: LogRecord) -> LogRecord | bool:  # type: ignore
        # copy so other handlers don't have a problem
        record = copy.copy(record)

        # update messages for copyability
        if not hasattr(record, "message"):
            record.message = record.msg % record.args

        record.exc_info = None
        record.exc_text = None
        record.stack_info = None

        self._send_message(
            LogChannel.STDLIB, record.name.encode(), pickle.dumps(record, pickle.HIGHEST_PROTOCOL)
        )

        return record

    def shutdown(self):
        self.socket.close()

    def send_structlog(self, logger, method, event_dict: EventDict):
        x = self._render(logger, method, {"method": method, "event": event_dict})
        if isinstance(x, str):
            x = x.encode()
        self._send_message(LogChannel.STRUCTLOG, logger.name.encode(), x)

        raise structlog.DropEvent()

    def send_task(self, task: Task):
        _log.debug("sending updated task", task_id=task.task_id)
        self._send_message(
            LogChannel.TASKS, str(task.task_id).encode(), task.model_dump_json().encode()
        )

    def send_progress(self, update: ProgressMessage):
        self._send_message(
            LogChannel.PROGRESS,
            str(update.progress_id).encode(),
            update.model_dump_json().encode(),
        )

    def send_record(self, sink_id: UUID, record: Any):
        self._send_message(LogChannel.RECORD, sink_id.hex.encode(), pickle.dumps(record))

    def _send_message(self, channel: LogChannel, name: bytes, data: bytes):
        mac = self._auth.hash_message(channel, name, data)

        with self._lock:
            self.socket.send_multipart([channel.value, name, data, mac])


def send_task(task: Task):
    if _active_context is not None:
        _active_context.send_task(task)
    else:
        _log.warn("no logging context is active")



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

# pyright: basic
from datetime import datetime

import numpy as np
import pandas as pd
import pyarrow as pa

from pytest import approx, mark, raises, warns

from lenskit.data import Dataset, DatasetBuilder
from lenskit.data.schema import AllowableTroolean
from lenskit.diagnostics import DataError, DataWarning


def test_add_repeated_interactions():
    dsb = DatasetBuilder()

    dsb.add_interactions(
        "click",
        pd.DataFrame(
            {
                "user_id": ["a", "a", "b", "c", "c", "c", "b"],
                "item_id": ["x", "y", "z", "x", "y", "z", "z"],
            }
        ),
        entities=["user", "item"],
        missing="insert",
    )

    ds = dsb.build()
    assert ds.user_count == 3
    assert ds.item_count == 3
    assert np.all(ds.users.ids() == ["a", "b", "c"])
    assert np.all(ds.items.ids() == ["x", "y", "z"])

    log = ds.interactions()
    assert log.schema.repeats == AllowableTroolean.PRESENT
    log = ds.interactions().pandas()
    assert isinstance(log, pd.DataFrame)
    assert all(log.columns == ["user_num", "item_num"])
    assert len(log) == 7


def test_remove_repeat_in_repeated_interactions():
    dsb = DatasetBuilder()
    dsb.add_relationship_class("click", ["user", "item"], allow_repeats=True, interaction=True)
    dsb.add_interactions(
        "click",
        pd.DataFrame(
            {
                "user_id": ["a", "a", "b", "c", "c", "c", "b"],
                "item_id": ["x", "y", "z", "x", "y", "z", "z"],
                "timestamp": np.arange(1, 8) * 10,
            }
        ),
        missing="insert",
        allow_repeats=True,
        remove_repeats=True,
    )

    ds = dsb.build()
    log = ds.interactions()
    assert log.schema.repeats.is_allowed
    log = ds.interactions().pandas()
    assert isinstance(log, pd.DataFrame)
    assert all(log.columns == ["user_num", "item_num", "timestamp"])
    assert len(log) == 6
    assert len(log[log.user_num == 1]) == 1


def test_remove_repeat_with_forbidden_repeat():
    dsb = DatasetBuilder()
    dsb.add_relationship_class("click", ["user", "item"], allow_repeats=False, interaction=True)
    dsb.add_interactions(
        "click",
        pd.DataFrame(
            {
                "user_id": ["a", "a", "b", "c", "c", "c", "b"],
                "item_id": ["x", "y", "z", "x", "y", "z", "z"],
                "timestamp": np.arange(1, 8) * 10,
            }
        ),
        missing="insert",
        allow_repeats=False,
        remove_repeats=True,
    )

    ds = dsb.build()
    log = ds.interactions()
    assert log.schema.repeats == AllowableTroolean.FORBIDDEN
    log = ds.interactions().pandas()
    assert isinstance(log, pd.DataFrame)
    assert all(log.columns == ["user_num", "item_num", "timestamp"])
    assert len(log) == 6
    assert len(log[log.user_num == 1]) == 1


def test_remove_duplicate_in_repeated_interactions():
    dsb = DatasetBuilder()
    dsb.add_relationship_class("click", ["user", "item"], allow_repeats=True, interaction=True)
    dsb.add_interactions(
        "click",
        pd.DataFrame(
            {
                "user_id": ["a", "a", "b", "c", "c", "c", "b"],
                "item_id": ["x", "y", "z", "x", "y", "y", "z"],
                "tag": ["j", "i", "k", "j", "k", "i", "k"],
            }
        ),
        missing="insert",
        remove_repeats="exact",
    )

    ds = dsb.build()
    log = ds.interactions().pandas()
    assert isinstance(log, pd.DataFrame)
    assert all(log.columns == ["user_num", "item_num", "tag"])
    assert len(log) == 6
    assert len(log[log.user_num == 1]) == 1
    assert len(log[log.user_num == 2]) == 3


def test_remove_repeat_interaction_at_builder_saves_last_interaction():
    dsb = DatasetBuilder()
    dsb.add_relationship_class("click", ["user", "item"], allow_repeats=True, interaction=True)
    dsb.add_interactions(
        "click",
        pd.DataFrame(
            {
                "user_id": ["a", "a", "b", "c", "c", "c", "b"],
                "item_id": ["x", "y", "z", "x", "y", "z", "z"],
                "timestamp": np.arange(1, 8) * 10,
                "rating": [3, 4, 5, 4, 3, 2, 3],
            }
        ),
        entities=["user", "item"],
        missing="insert",
        remove_repeats=True,
    )

    ds = dsb.build()
    log = ds.interactions().pandas()
    log = log[(log["user_num"] == 1) & (log["item_num"] == 2)]
    assert len(log) == 1
    assert log["rating"][5] == 3


def test_remove_repeat_interaction_at_matrix_saves_last_interaction():
    dsb = DatasetBuilder()
    dsb.add_relationship_class("click", ["user", "item"], allow_repeats=True, interaction=True)
    dsb.add_interactions(
        "click",
        pd.DataFrame(
            {
                "user_id": ["a", "a", "b", "c", "c", "c", "b"],
                "item_id": ["x", "y", "z", "x", "y", "z", "z"],
                "timestamp": np.arange(1, 8) * 10,
                "rating": [3, 4, 5, 4, 3, 2, 3],
            }
        ),
        entities=["user", "item"],
        missing="insert",
        allow_repeats=True,
    )

    ds = dsb.build()
    log = ds.interactions().matrix().pandas()
    log = log[(log["user_num"] == 1) & (log["item_num"] == 2)]
    assert len(log) == 1
    assert log["rating"][2] == 3


def test_remove_repeat_in_multiple_added_interactions():
    dsb = DatasetBuilder()
    dsb.add_relationship_class("click", ["user", "item"], allow_repeats=True, interaction=True)
    dsb.add_interactions(
        "click",
        pd.DataFrame(
            {
                "user_id": ["a", "a", "b", "c"],
                "item_id": ["x", "y", "z", "x"],
                "timestamp": np.arange(1, 5) * 10,
            }
        ),
        missing="insert",
        allow_repeats=True,
        remove_repeats=True,
    )
    dsb.add_interactions(
        "click",
        pd.DataFrame(
            {
                "user_id": ["c", "c", "b"],
                "item_id": ["y", "z", "z"],
                "timestamp": np.arange(5, 8) * 10,
            }
        ),
        missing="insert",
        allow_repeats=True,
        remove_repeats=True,
    )

    ds = dsb.build()
    log = ds.interactions()
    assert log.schema.repeats.is_allowed
    log = ds.interactions().pandas()
    assert isinstance(log, pd.DataFrame)
    assert all(log.columns == ["user_num", "item_num", "timestamp"])
    assert len(log) == 6
    assert len(log[log.user_num == 1]) == 1


def test_remove_duplicate_in_multiple_added_interactions():
    dsb = DatasetBuilder()
    dsb.add_relationship_class("click", ["user", "item"], allow_repeats=True, interaction=True)
    dsb.add_interactions(
        "click",
        pd.DataFrame(
            {
                "user_id": ["a", "a", "b", "c"],
                "item_id": ["x", "y", "z", "x"],
                "tag": ["j", "i", "k", "j"],
            }
        ),
        missing="insert",
        remove_repeats="exact",
    )
    dsb.add_interactions(
        "click",
        pd.DataFrame(
            {
                "user_id": ["c", "c", "b"],
                "item_id": ["y", "y", "z"],
                "tag": ["k", "i", "k"],
            }
        ),
        missing="insert",
        remove_repeats="exact",
    )

    ds = dsb.build()
    log = ds.interactions().pandas()
    assert isinstance(log, pd.DataFrame)
    assert all(log.columns == ["user_num", "item_num", "tag"])
    assert len(log) == 6
    assert len(log[log.user_num == 1]) == 1
    assert len(log[log.user_num == 2]) == 3


def test_bad_interaction_matrix_call():
    dsb = DatasetBuilder()
    with raises(ValueError):
        dsb.add_interactions(
            "click",
            pd.DataFrame(
                {
                    "user_id": ["a", "a", "b", "c", "c", "c", "b"],
                    "item_id": ["x", "y", "z", "x", "y", "z", "z"],
                }
            ),
            entities=["user", "item"],
            missing="insert",
        )
        ds = dsb.build()
        ds.interactions().matrix(row_entity="user", col_entity="user")


def test_repeated_interactions_timestamp():
    dsb = DatasetBuilder()

    dsb.add_interactions(
        "click",
        pd.DataFrame(
            {
                "user_id": ["a", "a", "b", "c", "c", "c", "b"],
                "item_id": ["x", "y", "z", "x", "y", "z", "z"],
                "timestamp": np.arange(1, 8) * 10,
            }
        ),
        entities=["user", "item"],
        missing="insert",
    )

    ds = dsb.build()

    log = ds.interactions().pandas()
    assert isinstance(log, pd.DataFrame)
    assert all(log.columns == ["user_num", "item_num", "timestamp"])
    assert len(log) == 7

    matrs = ds.interactions().matrix()
    mat = matrs.csr_structure()
    assert mat.nnz == 6
    assert np.all(mat.rowptrs == [0, 2, 3, 6])

    mat = matrs.scipy(attribute="timestamp")
    assert mat[1, 2] == 70

    mat = matrs.scipy(attribute="first_timestamp")
    assert mat[1, 2] == 30

    mat = matrs.scipy(attribute="count")
    assert mat[1, 2] == 2

    assert len(ds.user_row("a")) == 2
    assert len(ds.user_row("b")) == 1
    assert len(ds.user_row("c")) == 3


def test_repeated_interactions_count():
    dsb = DatasetBuilder()

    dsb.add_interactions(
        "click",
        pd.DataFrame(
            {
                "user_id": ["a", "a", "b", "c", "c", "c", "b"],
                "item_id": ["x", "y", "z", "x", "y", "z", "z"],
                "count": np.arange(1, 8),
            }
        ),
        entities=["user", "item"],
        missing="insert",
    )

    ds = dsb.build()

    matrs = ds.interactions().matrix()
    mat = matrs.csr_structure()
    assert mat.nnz == 6
    assert np.all(mat.rowptrs == [0, 2, 3, 6])

    mat = matrs.scipy(attribute="count")
    assert mat[1, 2] == 10


def test_repeated_interactions_count_with_null():
    dsb = DatasetBuilder()

    dsb.add_interactions(
        "click",
        pd.DataFrame(
            {
                "user_id": ["a", "a", "b", "c", "c", "c", "b"],
                "item_id": ["x", "y", "z", "x", "y", "z", "z"],
                "count": [1, 2, None, 4, 5, 6, 7],
            }
        ),
        entities=["user", "item"],
        missing="insert",
    )

    ds = dsb.build()

    matrs = ds.interactions().matrix()
    mat = matrs.csr_structure()
    assert mat.nnz == 6
    assert np.all(mat.rowptrs == [0, 2, 3, 6])

    mat = matrs.scipy(attribute="count")
    assert mat[1, 2] == 8


def test_add_three_entities_interactions():
    dsb = DatasetBuilder()

    dsb.add_interactions(
        "click",
        pd.DataFrame(
            {
                "user_id": ["a", "a", "b", "c", "c", "c", "b"],
                "item_id": ["x", "y", "z", "x", "y", "z", "z"],
                "tag_id": ["j", "i", "j", "j", "k", "i", "k"],
            }
        ),
        entities=["user", "item", "tag"],
        missing="insert",
    )

    ds = dsb.build()
    rset = ds.interactions()

    matrs_user_item = rset.matrix(row_entity="user", col_entity="item")
    mat = matrs_user_item.csr_structure()
    assert mat.nnz == 6
    assert np.all(mat.rowptrs == [0, 2, 3, 6])

    mat = matrs_user_item.scipy(attribute="count")
    assert mat[1, 2] == 2

    matrs_item_tag = rset.matrix(row_entity="item", col_entity="tag")
    mat = matrs_item_tag.csr_structure()
    assert mat.nnz == 6
    assert np.all(mat.rowptrs == [0, 1, 3, 6])

    mat = matrs_item_tag.scipy(attribute="count")
    assert mat[0, 1] == 2

    matrs_user_tag = rset.matrix(row_entity="user", col_entity="tag")
    mat = matrs_user_tag.csr_structure()
    assert mat.nnz == 7
    assert np.all(mat.rowptrs == [0, 2, 4, 7])

    mat = matrs_user_tag.scipy(attribute="count")
    assert np.all(mat.data == 1)


def test_matrix_relationship_set_cache():
    dsb = DatasetBuilder()

    dsb.add_interactions(
        "click",
        pd.DataFrame(
            {
                "user_id": ["a", "a", "b", "c", "c", "c", "b"],
                "item_id": ["x", "y", "z", "x", "y", "z", "z"],
            }
        ),
        entities=["user", "item"],
        missing="insert",
    )

    ds = dsb.build()
    matrix_user_item_a = ds.interactions().matrix(row_entity="user", col_entity="item")
    matrix_user_item_b = ds.interactions().matrix(row_entity="user", col_entity="item")

    assert matrix_user_item_a is matrix_user_item_b

    matrix_item_user = ds.interactions().matrix(row_entity="item", col_entity="user")

    assert matrix_item_user is not matrix_user_item_a



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
LensKit pipeline abstraction.
"""

from __future__ import annotations

from lenskit.diagnostics import PipelineError, PipelineWarning

from ._hooks import ComponentInputHook
from ._impl import CloneMethod, Pipeline
from ._profiling import PipelineProfiler, ProfileSink
from .builder import PipelineBuilder
from .cache import PipelineCache
from .common import RecPipelineBuilder, predict_pipeline, topn_pipeline
from .components import (
    Component,
    ComponentConstructor,
    PipelineFunction,
)
from .config import PipelineConfig
from .nodes import Node
from .state import PipelineState
from .types import Lazy

__all__ = [
    "Pipeline",
    "PipelineBuilder",
    "PipelineProfiler",
    "ProfileSink",
    "CloneMethod",
    "PipelineError",
    "PipelineWarning",
    "PipelineState",
    "Node",
    "PipelineFunction",
    "PipelineConfig",
    "Lazy",
    "Component",
    "ComponentConstructor",
    "PipelineCache",
    "RecPipelineBuilder",
    "topn_pipeline",
    "predict_pipeline",
    "ComponentInputHook",
]



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Abstraction for recording tasks.
"""

# pyright: strict
from __future__ import annotations

import socket
from enum import Enum
from os import PathLike
from pathlib import Path
from threading import Lock
from typing import Annotated, Any
from uuid import UUID, uuid4

import requests
from pydantic import AliasChoices, BaseModel, BeforeValidator, Field, SerializeAsAny
from typing_extensions import Literal

from ._proxy import get_logger
from .formats import friendly_duration
from .resource import ResourceMeasurement, reset_linux_hwm

_log = get_logger(__name__)
_active_tasks: list[Task] = []


def _dict_extract_values(data: object) -> Any:
    if isinstance(data, dict):
        return list(data.values())  # type: ignore
    else:
        return data


def _lk_machine():
    from lenskit.config import lenskit_config

    return lenskit_config().machine


class TaskStatus(str, Enum):
    """
    Statuses for task records.
    """

    PENDING = "pending"
    RUNNING = "running"
    FINISHED = "finished"
    FAILED = "failed"
    CANCELLED = "cancelled"


class Task(BaseModel, extra="allow"):
    """
    A task for logging and resource measurement.

    A task may be *top-level* (have no parent), or it may be a *subtask*.  By
    default, new tasks have the current active task as their parent.  Tasks are
    not active until they are started (using a task as a context manager
    automatically does this, which is the recommended process).

    The task-tracking mechanism is currently designed to support large tasks,
    like training a model or running batch inference; it is not yet designed for
    fine-grained spans like you would see in OpenTelemetry or Eliot.

    .. note::

        The notion of the “active task” does not yet support multi-threaded
        tasks.

    Stability:
        Caller

    Args:
        file:
            A file to save the task when it is finished.
        parent:
            The parent task.  If unspecified, uses the currently-active task.
        reset_hwm:
            Whether to reset the system resource high-water-marks at the start
            of this task.  Only effective on Linux, but allows for measurement
            of the peak memory use of this task specifically.  If unspecified,
            it resets the HWM if there is no parent.
    """

    task_id: UUID = Field(default_factory=uuid4, frozen=True)
    """
    The task ID.
    """
    parent_id: UUID | None = None
    """
    The parent task ID.
    """
    subprocess: bool = False
    """
    Whether this task is a subprocess of its parent.  Subprocess task CPU times
    are *not* included in the parent task times.
    """
    label: str
    """
    Human-readable task label.
    """
    hostname: str = Field(default_factory=socket.gethostname)
    """
    The hostname on which the task was run.
    """
    machine: str | None = Field(default_factory=_lk_machine)
    """
    The machine on which the task was run.

    Machines are project-meaningful labels for compute machines or clusters,
    primarily for understanding resource consumption.  See :ref:`settings`.
    """

    status: TaskStatus = TaskStatus.PENDING
    "The task's current status."
    start_time: float | None = None
    """
    The task start time (UNIX timestamp).
    """
    finish_time: float | None = None
    """
    The task completion time (UNIX timestamp).
    """

    duration: float | None = None
    """
    Task duration in seconds.  Measured using :func:`time.perf_counter`, so it
    may disagree slightly with the difference in start and finish times.
    """
    cpu_time: float | None = None
    """
    CPU time consumed in seconds.
    """
    peak_memory: int | None = None
    """
    Peak memory usage (max RSS) in bytes.  Only available on Unix; individual
    task peak memory use is only reliable on Linux (MacOS will report the max
    memory used since the process was started).
    """
    peak_gpu_memory: int | None = None
    """
    Peak PyTorch GPU memory usage in bytes.
    """

    system_power: float | None = Field(
        default=None, validation_alias=AliasChoices("system_power", "chassis_power")
    )
    """
    Estimated total system power consumption (in Joules).
    """
    cpu_power: float | None = None
    """
    Estimated CPU power consumption (in Joules).
    """
    gpu_power: float | None = None
    """
    Estimated GPU power consumption (in Joules).
    """

    subtasks: Annotated[list[SerializeAsAny[Task]], BeforeValidator(_dict_extract_values)] = Field(  # type: ignore
        default_factory=list
    )
    """
    This task's subtasks.
    """

    _reset_hwm: bool = False
    _save_file: Path | None = None
    _lock: Lock = None  # type: ignore
    _initial_meter: ResourceMeasurement | None = None
    _final_meter: ResourceMeasurement | None = None
    _refresh_id: UUID | None = None
    _subtask_index: dict[UUID, Task] | None = None

    @staticmethod
    def current() -> Task | None:
        """
        Get the currently-active task.
        """
        if _active_tasks:
            return _active_tasks[-1]

    @staticmethod
    def root() -> Task | None:
        """
        Get the root task.
        """
        if _active_tasks:
            return _active_tasks[0]

    def __init__(
        self,
        label: str,
        *,
        file: PathLike[str] | None = None,
        parent: Task | UUID | None = None,
        reset_hwm: bool | None = None,
        **data: Any,
    ):
        if isinstance(parent, Task):
            data["parent_id"] = parent.task_id
        elif parent is None and _active_tasks:
            data["parent_id"] = _active_tasks[-1].task_id
        super().__init__(label=label, **data)
        if reset_hwm is not None:
            self._reset_hwm = reset_hwm
        else:
            self._reset_hwm = parent is not None

        self._lock = Lock()

        if file:
            self._save_file = Path(file)

    def total_cpu(self) -> float | None:
        "Compute the total CPU time (including subprocesses)."
        time = self.cpu_time
        if time is None:
            return None

        for task in self.subtasks:
            if task.subprocess:
                child_time = task.total_cpu()
                if child_time is not None:
                    time += child_time

        return time

    @property
    def friendly_duration(self) -> str | None:
        if self.duration is None:
            return None
        else:
            return friendly_duration(self.duration)

    def save_to_file(self, path: PathLike[str], monitor: bool = True):
        """
        Save this task to a file, and re-save it when finished.
        """
        self._save_file = Path(path)
        self._save()

        # add to monitor if it isn't already
        if (
            monitor
            and self._initial_meter is not None
            and self._refresh_id is None
            and self._final_meter is None
        ):
            from lenskit.logging.monitor import get_monitor

            mon = get_monitor()
            self._refresh_id = mon.add_refreshable(self)

    def start(self):
        """
        Start the task.
        """
        if self._reset_hwm:
            reset_linux_hwm()

        log = _log.bind(task_id=self.task_id)
        self._initial_meter = ResourceMeasurement.current()
        self.start_time = self._initial_meter.wall_time
        self.status = TaskStatus.RUNNING
        log.debug("beginning task")
        self._save()

        if self.parent_id:
            cur = self.current()
            if cur is not None:
                if cur.task_id == self.parent_id:
                    cur.add_subtask(self)
                else:
                    log.warn("active task is not parent")
            else:
                log.debug("have a task but no parent")

        _active_tasks.append(self)
        if self._save_file:
            from lenskit.logging.monitor import get_monitor

            mon = get_monitor()
            self._refresh_id = mon.add_refreshable(self)

    def finish(self, status: TaskStatus = TaskStatus.FINISHED):
        """
        Finish the task.
        """
        log = _log.bind(task_id=str(self.task_id), label=self.label)
        if _active_tasks[-1] is not self:
            log.warn("attempted to finish non-active task")
            _active_tasks.remove(self)
        else:
            _active_tasks.pop()

        with self._lock:
            if self._refresh_id is not None:
                from lenskit.logging.monitor import get_monitor

                mon = get_monitor()
                mon.remove_refreshable(self._refresh_id)

            self._final_meter = self.update_resources()
            self.finish_time = self._final_meter.wall_time
            self.status = status
            log.debug("finished task", time=self.duration, cpu=self.cpu_time)
            self._save()

    def update(self):
        """
        Update the task's resource measurements and save the file (if one is set).
        """
        with self._lock:
            self.update_resources()
            self._save()

    def monitor_refresh(self):
        """
        Refresh method called by the monitor backend.
        """
        self.update()

    def add_subtask(self, task: Task):
        """
        Add or update a subtask.
        """
        _log.debug("adding subtask", task=task.model_dump())
        with self._lock:
            if self._subtask_index is None:
                self._subtask_index = {t.task_id: t for t in self.subtasks}
            self._subtask_index[task.task_id] = task

            if task.parent_id is None:
                task.parent_id = self.task_id

            self.subtasks = list(self._subtask_index.values())

    def _save(self):
        if self._save_file:
            with self._save_file.open("w") as f:
                print(self.model_dump_json(), file=f)

    def update_resources(self) -> ResourceMeasurement:
        """
        Update the resource measurements.  Returns the current measurement.

        This method is called by :meth:`update`, with an exclusive lock held.
        """
        assert self._initial_meter is not None
        now = ResourceMeasurement.current()
        elapsed = now - self._initial_meter
        self.duration = elapsed.perf_time
        self.cpu_time = elapsed.cpu_time
        self.peak_memory = elapsed.max_rss
        self.peak_gpu_memory = elapsed.max_gpu

        self.system_power = measure_power("system", elapsed.perf_time)
        self.cpu_power = measure_power("cpu", elapsed.perf_time)
        self.gpu_power = measure_power("gpu", elapsed.perf_time)

        return now

    def __enter__(self):
        self.start()
        return self

    def __exit__(
        self, exc_type: type[Exception] | None, exc_value: Exception | None, traceback: Any
    ):
        log = _log.bind(task_id=str(self.task_id), label=self.label)
        if exc_type is None:
            status = TaskStatus.FINISHED
        elif issubclass(exc_type, KeyboardInterrupt):
            status = TaskStatus.CANCELLED
            log.debug("task cancelled")
        else:
            status = TaskStatus.FAILED
            log.debug("task failed: %s", exc_value)

        self.finish(status)

    def __reduce__(self):
        return self.__class__.model_validate, (self.model_dump(),)

    def __str__(self):
        return f"<Task {self.task_id}: {self.label}>"


def measure_power(scope: Literal["system", "cpu", "gpu"], duration: float):
    from lenskit.config import lenskit_config

    time_ms = int(duration * 1000)

    config = lenskit_config()
    prom = config.prometheus.url
    if not prom:
        return None

    machine = config.current_machine
    if machine is None:
        return None

    url = prom + "/api/v1/query"
    if query := machine.power_queries.get(scope):
        query = query.format(elapsed=time_ms, machine=config.machine)
        return _get_prometheus_metric(url, query, time_ms)


def _get_prometheus_metric(url: str, query: str, time_ms: int) -> float | None:
    log = _log.bind(url=url, query=query)
    try:
        res = requests.get(url, {"query": query}).json()
    except Exception as e:
        log.warning("Prometheus query error", exc_info=e)
        return None

    log.debug("received response", response=res)
    if res["status"] == "error":
        log.error("Prometheus query error: %s", res["error"], type=res["errorType"])
        return None

    results = res["data"]["result"]
    if len(results) == 0:
        log.debug("Prometheus query returned no results")
        return None
    elif len(results) > 1:
        log.error("Prometheus query returned %d results, expected 1", len(results))

    _time, val = results[0]["value"]
    return float(val)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import numpy as np
from typing_extensions import override

from lenskit.data.items import ItemList

from ._base import ListMetric, RankingMetricBase
from ._weighting import GeometricRankWeight, RankWeight


class RBP(ListMetric, RankingMetricBase):
    """
    Evaluate recommendations with rank-biased precision :cite:p:`rbp`.

    If :math:`r_{ui} \\in \\{0, 1\\}` is binary implicit ratings, and the
    weighting is the default geometric weight with patience :math:`p`, the RBP
    is computed by:

    .. math::
        \\begin{align*}
        \\operatorname{RBP}_p(L, u) & =(1 - p) \\sum_i r_{ui} p^i
        \\end{align*}

    The original RBP metric depends on the idea that the rank-biased sum of
    binary relevance scores in an infinitely-long, perfectly-precise list has is
    :math:`1/(1 - p)`.  If RBP is used with a non-standard weighting that does
    not have a defined infinite series sum, then this metric will normalize by
    the sum of the discounts for the recommendation list.

    :cite:t:`rbp` provide an extended discussion on choosing the patience
    parameter :math:`\\gamma`.  This metric defaults to :math:`\\gamma=0.85`, to
    provide a relatively shallow curve and reward good items on the first few
    pages of results (in a 10-per-page setting).  Recommendation systems data
    has no pooling, so the variance of this estimator may be high as they note
    in the paper; however, RBP with high patience should be no worse than nDCG
    (and perhaps even better) in this regard.

    In recommender evaluation, we usually have a small test set, so the maximum
    achievable RBP is significantly less than the theoretical maximum, and is a
    function of the number of test items.  With ``normalize=True``, the RBP
    metric will be normalized by the maximum achievable with the provided test
    data, like NDCG.

    .. warning::

        The additional normalization is experimental, and should not yet be used
        for published research results.

    Args:
        k:
            The maximum recommendation list length.
        weight:
            The rank weighting model to use.  Defaults to
            :class:`GeometricRankWeight` with the specified patience parameter.
        patience:
            The patience parameter :math:`p`, the probability that the user
            continues browsing at each point.  The default is 0.85.
        normalize:
            Whether to normalize the RBP scores; if ``True``, divides the RBP
            score by the maximum achievable with the test data (as in nDCG).

    Stability:
        Caller
    """

    weight: RankWeight
    patience: float
    normalize: bool

    def __init__(
        self,
        k: int | None = None,
        *,
        weight: RankWeight | None = None,
        patience: float = 0.85,
        normalize: bool = False,
    ):
        super().__init__(k)
        self.patience = patience
        if weight is None:
            weight = GeometricRankWeight(patience)
        self.weight = weight
        self.normalize = normalize

    @property
    def label(self):
        if self.k is not None:
            return f"RBP@{self.k}"
        else:
            return "RBP"

    @override
    def measure_list(self, recs: ItemList, test: ItemList) -> float:
        recs = self.truncate(recs)
        k = len(recs)

        nrel = len(test)
        if nrel == 0:
            return np.nan

        good = recs.isin(test)

        ranks = recs.ranks()
        assert ranks is not None

        # we only need to sum good weights
        good_ranks = ranks[good]
        weight = self.weight.weight(good_ranks)
        rbp = np.sum(weight).item()

        # figure out normalization
        wmax = self.weight.series_sum()
        if self.normalize:
            # normalize by max achieveable RBP
            max = np.sum(weight[: min(nrel, k)]).item()
            return rbp / max
        elif wmax is not None:
            # normalization defined by metric
            return rbp / wmax
        else:
            # normalize by total weight
            all_weight = self.weight.weight(ranks)
            return rbp / np.sum(all_weight).item()



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Tests for the pipeline type-checking functions.
"""

import typing
from collections.abc import Iterable, Sequence
from pathlib import Path
from types import NoneType
from typing import Any, TypeVar
from uuid import UUID

import numpy as np
import pandas as pd
from numpy.typing import ArrayLike, NDArray

from pytest import warns

from lenskit.data import Dataset, MatrixRelationshipSet, RelationshipSet
from lenskit.pipeline.types import (
    TypecheckWarning,
    is_compatible_data,
    is_compatible_type,
    resolve_type_string,
    type_string,
)

T = TypeVar("T")
Tstr = TypeVar("Tstr", bound=str)


def test_type_compat_identical():
    assert is_compatible_type(int, int)
    assert is_compatible_type(str, str)


def test_type_compat_subclass():
    assert is_compatible_type(MatrixRelationshipSet, RelationshipSet)


def test_type_compat_assignable():
    assert is_compatible_type(int, float)


def test_type_raw_compat_with_generic():
    assert is_compatible_type(list, list[int])
    assert not is_compatible_type(set, list[int])


def test_type_compat_protocol():
    assert is_compatible_type(list, Sequence)
    assert is_compatible_type(list, typing.Sequence)
    assert not is_compatible_type(set, Sequence)
    assert not is_compatible_type(set, typing.Sequence)
    assert is_compatible_type(set, Iterable)


def test_type_compat_protocol_generic():
    assert is_compatible_type(list, Sequence[int])
    assert is_compatible_type(list, typing.Sequence[int])


def test_type_compat_generics_with_protocol():
    assert is_compatible_type(list[int], Sequence[int])


def test_type_incompat_generics():
    with warns(TypecheckWarning):
        assert is_compatible_type(list[int], list[str])
    with warns(TypecheckWarning):
        assert is_compatible_type(list[int], Sequence[str])


def test_data_compat_basic():
    assert is_compatible_data(72, int)
    assert is_compatible_data("hello", str)
    assert not is_compatible_data(72, str)


def test_data_compat_float_assignabile():
    assert is_compatible_data(72, float)


def test_data_compat_generic():
    assert is_compatible_data(["foo"], list[str])
    # this is compatible because we can't check generics
    # with warns(TypecheckWarning):
    assert is_compatible_data([72], list[str])


def test_numpy_typecheck():
    assert is_compatible_data(np.arange(10, dtype="i8"), NDArray[np.int64])
    assert is_compatible_data(np.arange(10, dtype="i4"), NDArray[np.int32])
    assert is_compatible_data(np.arange(10), ArrayLike)
    assert is_compatible_data(np.arange(10), NDArray[np.integer])
    # numpy types can be checked
    assert not is_compatible_data(np.arange(10), NDArray[np.float64])


def test_numpy_scalar_typecheck():
    assert is_compatible_data(np.int32(4270), np.integer[Any])


def test_numpy_scalar_typecheck2():
    assert is_compatible_data(np.int32(4270), np.integer[Any] | int)


def test_pandas_typecheck():
    assert is_compatible_data(pd.Series(["a", "b"]), ArrayLike)


def test_compat_with_typevar():
    assert is_compatible_data(100, T)


def test_not_compat_with_typevar():
    assert not is_compatible_data(100, Tstr)


def test_type_string_none():
    assert type_string(None) == "None"


def test_type_string_str():
    assert type_string(str) == "str"


def test_type_string_generic():
    assert type_string(list[str]) == "list"


def test_type_string_class():
    assert type_string(UUID) == "uuid.UUID"


def test_parse_string_None():
    assert resolve_type_string("None") == NoneType


def test_parse_string_int():
    assert resolve_type_string("int") is int


def test_parse_string_class():
    assert resolve_type_string("pathlib.Path") is Path


def test_parse_string_mod_class():
    assert resolve_type_string("pathlib:Path") is Path



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import pickle

import pandas as pd

from lenskit.basic import PopScorer
from lenskit.data.items import ItemList
from lenskit.testing import BasicComponentTests, ScorerTests

simple_df = pd.DataFrame(
    {"item": [1, 1, 2, 3], "user": [10, 12, 10, 13], "rating": [4.0, 3.0, 5.0, 2.0]}
)


class TestPopScore(BasicComponentTests, ScorerTests):
    component = PopScorer

    def verify_models_equivalent(self, orig, copy):
        assert all(copy.item_scores_ == orig.item_scores_)


def test_popscore_quantile(rng, ml_ds):
    pop = PopScorer()
    pop.train(ml_ds)

    assert pop.item_scores_.max() == 1.0

    counts = ml_ds.item_stats()["count"]
    counts = counts.sort_values()

    winner = counts.index[-1]
    assert pop.item_scores_[ml_ds.items.number(winner)] == 1.0


def test_popscore_rank(rng, ml_ds):
    pop = PopScorer(score="rank")
    pop.train(ml_ds)

    counts = ml_ds.item_stats()["count"]
    counts = counts.sort_values()

    assert pop.item_scores_.max() == len(counts)

    winner = counts.index[-1]
    assert pop.item_scores_[ml_ds.items.number(winner)] == len(counts)


def test_popscore_counts(rng, ml_ds):
    pop = PopScorer(score="count")
    pop.train(ml_ds)

    counts = ml_ds.item_stats()["count"]

    scores, counts = pd.Series(pop.item_scores_, index=pop.items_.ids()).align(counts)
    assert all(scores == counts)

    items = rng.choice(counts.index, 100)
    scores = pop(ItemList(item_ids=items))
    assert all(scores.scores() == counts.loc[items])



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import logging

import numpy as np
import torch

import hypothesis.strategies as st
from hypothesis import given
from pytest import approx

from lenskit.math.solve import solve_cholesky

_log = logging.getLogger(__name__)


@st.composite
def square_problem(draw, scale=10):
    size = draw(st.integers(2, 100))

    # Hypothesis doesn't do well at generating problem data, so go with this
    seed = draw(st.integers(min_value=0, max_value=2**32 - 1))
    rng = np.random.RandomState(seed)
    A = rng.randn(size, size) * scale
    b = rng.randn(size) * scale
    return A, b, size


@given(square_problem())
def test_solve_cholesky(problem):
    A, b, size = problem

    # square values of A so it's positive
    A = A * A

    # and solve least squares
    xexp, resid, rank, s = np.linalg.lstsq(A, b, rcond=None)

    F = A.T @ A
    y = A.T @ b
    x = solve_cholesky(F, y)
    assert x.shape == y.shape

    assert x == approx(xexp, rel=1.0e-3)

    assert F @ x == approx(y, rel=2.0e-6, abs=5.0e-9)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
LensKit reranking metrics.
"""

from ._lip import least_item_promoted
from ._rbo import rank_biased_overlap

__all__ = ["least_item_promoted", "rank_biased_overlap"]



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import json
from os import fspath
from pathlib import Path

from click.testing import CliRunner

from lenskit.cli import lenskit
from lenskit.data import Dataset, ItemListCollection


def test_doctor():
    runner = CliRunner()
    result = runner.invoke(lenskit, ["doctor"])

    assert result.exit_code == 0


def test_doctor_full():
    runner = CliRunner()
    result = runner.invoke(lenskit, ["doctor", "--packages", "--paths"])

    assert result.exit_code == 0



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from uuid import UUID, uuid4

import numpy as np
import pandas as pd
import pyarrow as pa
import pyarrow.compute as pc

from pytest import mark

from lenskit import _accel

VOCAB_SIZE = 50000


@mark.benchmark(group="single")
def test_pandas_integer_single(rng: np.random.Generator, benchmark):
    ids = rng.choice(VOCAB_SIZE * 1000, VOCAB_SIZE, replace=False)
    index = pd.Index(ids)

    query = rng.choice(ids, 1).item()

    def search():
        _val = index.get_loc(query)

    benchmark(search)


@mark.benchmark(group="array")
def test_pandas_integer_500(rng: np.random.Generator, benchmark):
    ids = rng.choice(VOCAB_SIZE * 1000, VOCAB_SIZE, replace=False)
    index = pd.Index(ids)

    query = rng.choice(ids, 500, replace=False)

    def search():
        _val = index.get_indexer_for(query)

    benchmark(search)


@mark.benchmark(group="single")
def test_pandas_string_single(rng: np.random.Generator, benchmark):
    ids = [str(uuid4()) for i in range(VOCAB_SIZE)]
    index = pd.Index(ids)

    query = rng.choice(ids, 1).item()

    def search():
        _val = index.get_loc(query)

    benchmark(search)


@mark.benchmark(group="array")
def test_pandas_string_500(rng: np.random.Generator, benchmark):
    ids = [str(uuid4()) for i in range(VOCAB_SIZE)]
    index = pd.Index(ids)

    query = rng.choice(ids, 500, replace=False)

    def search():
        _val = index.get_indexer_for(query)

    benchmark(search)


@mark.benchmark(group="single")
def test_arrow_integer_single(rng: np.random.Generator, benchmark):
    ids = rng.choice(VOCAB_SIZE * 1000, VOCAB_SIZE, replace=False)
    index = pa.array(ids)

    query = rng.choice(ids, 1).item()

    def search():
        _val = index.index(query)

    benchmark(search)


@mark.benchmark(group="array")
def test_arrow_integer_500(rng: np.random.Generator, benchmark):
    ids = rng.choice(VOCAB_SIZE * 1000, VOCAB_SIZE, replace=False)
    index = pa.array(ids)

    query = rng.choice(ids, 500)

    def search():
        _val = pc.index_in(query, index)

    benchmark(search)


@mark.benchmark(group="single")
def test_arrow_string_single(rng: np.random.Generator, benchmark):
    ids = [str(uuid4()) for i in range(VOCAB_SIZE)]
    index = pa.array(ids)

    query = rng.choice(ids, 1).item()

    def search():
        _val = index.index(query)

    benchmark(search)


@mark.benchmark(group="array")
def test_arrow_string_500(rng: np.random.Generator, benchmark):
    ids = [str(uuid4()) for i in range(VOCAB_SIZE)]
    index = pa.array(ids)

    query = rng.choice(ids, 500)

    def search():
        _val = pc.index_in(query, index)

    benchmark(search)


@mark.benchmark(group="single")
def test_rust_integer_single(rng: np.random.Generator, benchmark):
    ids = rng.choice(VOCAB_SIZE * 1000, VOCAB_SIZE, replace=False)
    index = _accel.data.IDIndex(pa.array(ids))

    query = rng.choice(ids, 1).item()

    def search():
        _val = index.get_index(query)

    benchmark(search)


@mark.benchmark(group="array")
def test_rust_integer_500(rng: np.random.Generator, benchmark):
    ids = rng.choice(VOCAB_SIZE * 1000, VOCAB_SIZE, replace=False)
    index = _accel.data.IDIndex(pa.array(ids))

    query = pa.array(rng.choice(ids, 500, replace=False))

    def search():
        _val = index.get_indexes(query)

    benchmark(search)


@mark.benchmark(group="single")
def test_rust_string_single(rng: np.random.Generator, benchmark):
    ids = [str(uuid4()) for i in range(VOCAB_SIZE)]
    index = _accel.data.IDIndex(pa.array(ids))

    query = rng.choice(ids, 1).item()

    def search():
        _val = index.get_index(query)

    benchmark(search)


@mark.benchmark(group="array")
def test_rust_string_500(rng: np.random.Generator, benchmark):
    ids = [str(uuid4()) for i in range(VOCAB_SIZE)]
    index = _accel.data.IDIndex(pa.array(ids))

    query = pa.array(rng.choice(ids, 500, replace=False))

    def search():
        _val = index.get_indexes(query)

    benchmark(search)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from dataclasses import dataclass
from typing import Any

import numpy as np
import ray.tune
import ray.tune.result

from lenskit.logging import Progress, get_logger, item_progress

_log = get_logger("codex.tuning")


@dataclass
class TrialProgress:
    bar: Any
    count: int


class StatusCallback(ray.tune.Callback):
    def __init__(self, model: str | None, ds: str | None):
        self.log = _log.bind(model=model, dataset=ds)

    def on_trial_result(self, iteration, trials, trial, result, **info):
        metrics = {n: v for (n, v) in result.items() if n in ["RBP", "NDCG", "RecipRank", "RMSE"]}
        self.log.info("new trial result", iter=iteration, id=trial.trial_id, **metrics)


class ProgressReport(ray.tune.ProgressReporter):
    job_name: str | None = None
    metric = None
    mode = None
    best_metric = None
    _bar: Progress
    _task_bars: dict[str, TrialProgress]

    def __init__(self, name: str | None = None):
        super().__init__()
        self.job_name = name
        self.done = set()

    def setup(self, start_time=None, total_samples=None, metric=None, mode=None, **kwargs):
        super().setup(start_time, total_samples, metric, mode, **kwargs)

        _log.info("setting up tuning status", total_samples=total_samples, metric=metric, mode=mode)
        extra = {metric: ".3f"} if metric is not None else {}
        label = "Tuning " + (self.job_name or "trials")
        self._bar = item_progress(label, total_samples, extra)
        self._task_bars = {}
        self.metric = metric
        self.mode = mode

    def report(self, trials, done, *sys_info):
        _log.debug("reporting trial completion", trial_count=len(trials))

        if done:
            _log.debug("search complete", trial_count=len(trials))
            for bar in self._task_bars.values():
                bar.bar.finish()
            self._bar.finish()
        else:
            total = len(trials)
            if total <= self._bar.total:
                total = None

            n_new = 0
            for trial in trials:
                self._update_metric(trial)
                if trial.status == "TERMINATED" and trial.trial_id not in self.done:
                    self.done.add(trial.trial_id)
                    n_new += 1
                    _log.debug("finished trial", id=trial.trial_id, config=trial.config)

                if trial.status != "RUNNING" and trial.trial_id in self._task_bars:
                    self._task_bars[trial.trial_id].bar.finish()
                    del self._task_bars[trial.trial_id]

                if (
                    trial.status == "RUNNING"
                    and trial.last_result
                    and "training_iteration" in trial.last_result
                ):
                    tp = self._task_bars.get(trial.trial_id, None)
                    epoch = trial.last_result["training_iteration"]
                    t_total = trial.last_result.get("max_epochs", None)
                    if tp is None:
                        bar = item_progress(
                            "Trial " + trial.trial_id,
                            total=t_total,
                            fields={self.metric: ".3f"},
                        )
                        tp = TrialProgress(bar, 0)
                        self._task_bars[trial.trial_id] = tp

                    if epoch > tp.count:
                        tp.bar.update(
                            epoch - tp.count,
                            total=t_total,
                            **{self.metric: trial.last_result[self.metric]},
                        )
                        tp.count = epoch

            extra = {self.metric: self.best_metric or np.nan}
            self._bar.update(n_new, total=total, **extra)

    def should_report(self, trials, done=False):
        return True

    def _update_metric(self, trial):
        if self.metric is not None and trial.last_result and self.metric in trial.last_result:
            mv = trial.last_result[self.metric]
            if self.best_metric is None:
                self.best_metric = mv
                return True
            elif self.mode == "max" and mv > self.best_metric:
                self.best_metric = mv
                return True
            elif self.mode == "min" and mv < self.best_metric:
                self.best_metric = mv
                return True

        return False



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Code to import MovieLens data sets into LensKit.
"""

from __future__ import annotations

import re
from collections.abc import Callable
from io import TextIOWrapper
from pathlib import Path
from typing import TypeAlias
from zipfile import ZipFile

import numpy as np
import pandas as pd
import structlog
from scipy.sparse import coo_array

from lenskit.logging import get_logger

from .builder import DatasetBuilder
from .dataset import Dataset

_log = get_logger(__name__)

LOC: TypeAlias = Path | tuple[ZipFile, str]


class MLData:
    """
    Internal class representing an open ML data set.

    .. stability:: internal
    """

    version: str
    source: Path | ZipFile
    prefix: str = ""

    _logger: structlog.stdlib.BoundLogger

    def __init__(self, version: str, source: Path | ZipFile, prefix: str = ""):
        self.version = version
        self.source = source
        self.prefix = prefix

        log = _log.bind(version=version)
        if isinstance(source, ZipFile):
            log = log.bind(source=source.filename)
        else:
            log = log.bind(source=str(source))
        self._logger = log

    @staticmethod
    def version_impl(version: str) -> Callable[..., MLData]:
        if version == "ml-100k":
            return ML100KLoader
        elif re.match(r"^ml-10?m(100k)?$", version, re.IGNORECASE):
            return MLMLoader
        elif re.match(r"^ml-(\d+m|latest(-small)?)$", version, re.IGNORECASE):
            return MLModernLoader
        else:  # pragma: nocover
            raise ValueError(f"unknown ML version {version}")

    def __enter__(self):
        return self

    def __exit__(self, *args):
        if isinstance(self.source, ZipFile):
            self.source.close()

    def open_file(self, name: str, encoding: str = "utf8"):
        if isinstance(self.source, Path):
            return open(self.source / (self.prefix + name), "rb")
        else:
            self._logger.debug("opening zip archive member")
            return self.source.open(self.prefix + name)

    def dataset(self) -> Dataset:  # pragma: nocover
        """
        Load the full dataset.
        """
        raise NotImplementedError()

    def ratings_df(self) -> pd.DataFrame:  # pragma: nocover
        """
        Load the ratings data frame.
        """
        raise NotImplementedError()


class ML100KLoader(MLData):
    """
    Loader for the ML100K data set.
    """

    def dataset(self) -> Dataset:
        dsb = DatasetBuilder(self.version)

        genres = self.genres().tolist()
        movies = self.movies_df(genres)
        movies = movies.drop(columns=["misc"])

        dsb.add_entities("item", movies["item_id"])
        dsb.add_scalar_attribute("item", "title", movies["item_id"], movies["title"])
        # TODO: add movie genres

        ratings = self.ratings_df()
        dsb.add_interactions(
            "rating",
            ratings,
            entities=["user", "item"],
            missing="insert",
            default=True,
            allow_repeats=False,
        )

        users = self.users_df().set_index("user_id")
        dsb.add_entities("user", users.index.values, duplicates="update")
        for c in users.columns:
            self._logger.debug("adding user column %s", c)
            dsb.add_scalar_attribute("user", c, users[c])

        return dsb.build()

    def genres(self) -> pd.Series:
        self._logger.debug("reading ML100K genre list")
        with self.open_file("u.genre") as data:
            df = pd.read_csv(data, sep="|", names=["name", "number"])
            return df.set_index("number").sort_index()["name"]

    def movies_df(self, genres: list[str] | None = None) -> pd.DataFrame:
        if genres is None:
            genres = self.genres().tolist()
        self._logger.debug("reading ML100K movie info")
        with self.open_file("u.item") as data:
            return pd.read_csv(
                data,
                sep=r"\|",
                header=None,
                names=["item_id", "title", "date", "misc", "IMDB"] + genres,
                dtype={
                    "item_id": np.int32,
                },
                encoding="latin1",
            )

    def users_df(self) -> pd.DataFrame:
        self._logger.debug("reading ML100K user info")
        with self.open_file("u.user") as data:
            return pd.read_csv(
                data,
                sep=r"\|",
                header=None,
                names=["user_id", "age", "gender", "occupation", "zip_code"],
                dtype={
                    "user_id": np.int32,
                },
                encoding="latin1",
            )

    def ratings_df(self) -> pd.DataFrame:
        self._logger.debug("reading ML100K ratings TSV")
        with self.open_file("u.data") as data:
            df = pd.read_csv(
                data,
                sep="\t",
                header=None,
                names=["user_id", "item_id", "rating", "timestamp"],
                dtype={
                    "user_id": np.int32,
                    "item_id": np.int32,
                    "rating": np.float32,
                    "timestamp": np.int32,
                },
            )
            df["timestamp"] = pd.to_datetime(df["timestamp"], unit="s")
            return df


class MLMLoader(MLData):
    """
    Loader for the ML 1M and 10M data sets.
    """

    def dataset(self) -> Dataset:
        dsb = DatasetBuilder(self.version)

        movies = self.movies_df()

        dsb.add_entities("item", movies["item_id"])
        dsb.add_scalar_attribute("item", "title", movies["item_id"], movies["title"])
        genre_strings = movies["genres"]
        genres = movies["genres"].str.split("|", regex=False)
        genres[genre_strings == "(no genres listed)"] = None
        dsb.add_list_attribute("item", "genres", movies["item_id"], genres, dictionary=True)

        ratings = self.ratings_df()
        dsb.add_interactions(
            "rating",
            ratings,
            entities=["user", "item"],
            missing="insert",
            default=True,
            allow_repeats=False,
        )

        users = self.users_df()
        if users is not None:
            users = users.set_index("user_id")
            dsb.add_entities("user", users.index.values, duplicates="update")
            for c in users.columns:
                self._logger.debug("adding user column %s", c)
                dsb.add_scalar_attribute("user", c, users[c])

        tags = self.tagging_df()
        if tags is not None:
            # we add them as a sparse vector attribute of counts
            tag_names = np.unique(tags["tag"])
            tag_idx = pd.Index(tag_names)
            tags["tag_num"] = tag_idx.get_indexer_for(tags["tag"])
            tag_counts = (
                tags.groupby(["item_id", "tag_num"])["tag"].count().reset_index(name="count")
            )
            tag_items = np.unique(tags["item_id"])
            item_idx = pd.Index(tag_items)
            icol = item_idx.get_indexer_for(tag_counts["item_id"])
            tag_matrix = coo_array((tag_counts["count"], (icol, tag_counts["tag_num"])))
            dsb.add_vector_attribute("item", "tag_counts", tag_items, tag_matrix)

        return dsb.build()

    def movies_df(self):
        self._logger.debug("reading ML10?M movies file")
        with self.open_file("movies.dat") as data:
            return pd.read_csv(
                data,
                sep=":",
                header=None,
                names=["item_id", "_it", "title", "_tg", "genres"],
                usecols=[0, 2, 4],
                dtype={
                    "item_id": np.int32,
                },
                encoding="latin1",
            )

    def users_df(self):
        if self.version != "ml-1m":
            return None

        self._logger.debug("reading ML1M users file")
        with self.open_file("users.dat") as data:
            return pd.read_csv(
                data,
                sep=":",
                header=None,
                names=[
                    "user_id",
                    "_ug",
                    "gender",
                    "_ga",
                    "age",
                    "_ao",
                    "occupation",
                    "_az",
                    "zip_code",
                ],
                usecols=[0, 2, 4, 6, 8],
                dtype={
                    "user_id": np.int32,
                },
                encoding="latin1",
            )

    def ratings_df(self):
        self._logger.debug("reading ML10?M ratings file")
        with self.open_file("ratings.dat") as data:
            df = pd.read_csv(
                data,
                sep=":",
                header=None,
                names=["user_id", "_ui", "item_id", "_ir", "rating", "_rt", "timestamp"],
                usecols=[0, 2, 4, 6],
                dtype={
                    "user_id": np.int32,
                    "item_id": np.int32,
                    "rating": np.float32,
                    "timestamp": np.int32,
                },
            )
            df["timestamp"] = pd.to_datetime(df["timestamp"], unit="s")
            return df

    def tagging_df(self):
        if self.version != "ml-10m":
            return None

        self._logger.debug("reading ML10M tags file")
        # this is slow but the data is a mess
        lpat = re.compile(r"^(\d+)::(\d+)::(.*)::(\d+)$")
        lines = []
        with self.open_file("tags.dat") as data:
            for i, line in enumerate(TextIOWrapper(data, "latin1"), 1):
                m = lpat.match(line)
                if not m:  # pragma: nocover
                    self._logger.warn("invalid line", line=i)
                    continue

                lines.append((int(m[1]), int(m[2]), m[3], int(m[4])))

        df = pd.DataFrame.from_records(lines, columns=["user_id", "item_id", "tag", "timestamp"])
        df["timestamp"] = pd.to_datetime(df["timestamp"], unit="s")
        return df


class MLModernLoader(MLData):
    """
    Loader for modern MovieLens data sets (20M and later).
    """

    def dataset(self) -> Dataset:
        dsb = DatasetBuilder(self.version)

        movies = self.movies_df()

        dsb.add_entities("item", movies["item_id"])
        dsb.add_scalar_attribute("item", "title", movies["item_id"], movies["title"])
        genre_strings = movies["genres"]
        genres = movies["genres"].str.split("|", regex=False)
        genres[genre_strings == "(no genres listed)"] = None
        dsb.add_list_attribute("item", "genres", movies["item_id"], genres)

        ratings = self.ratings_df()
        dsb.add_interactions(
            "rating",
            ratings,
            entities=["user", "item"],
            missing="insert",
            default=True,
            allow_repeats=False,
        )

        tags = self.tagging_df()
        # we add them as a sparse vector attribute of counts
        tag_names = np.unique(tags["tag"])
        tag_idx = pd.Index(tag_names)
        tags["tag_num"] = tag_idx.get_indexer_for(tags["tag"])
        tag_counts = tags.groupby(["item_id", "tag_num"])["tag"].count().reset_index(name="count")
        tag_items = np.unique(tags["item_id"])
        item_idx = pd.Index(tag_items)
        icol = item_idx.get_indexer_for(tag_counts["item_id"])
        tag_matrix = coo_array((tag_counts["count"], (icol, tag_counts["tag_num"])))
        dsb.add_vector_attribute("item", "tag_counts", tag_items, tag_matrix)

        genome = self.genome_df()
        if genome is not None:
            dsb.add_vector_attribute(
                "item",
                "tag_genome",
                genome.index.values,
                genome.to_numpy(),
                dim_names=genome.columns,
            )

        return dsb.build()

    def movies_df(self):
        self._logger.debug("reading modern movies CSV")
        with self.open_file("movies.csv") as data:
            df = pd.read_csv(
                data,
                dtype={
                    "movieId": np.int32,
                },
            ).rename(columns={"movieId": "item_id"})
            return df

    def tagging_df(self):
        self._logger.debug("reading modern movies CSV")
        with self.open_file("tags.csv") as data:
            df = pd.read_csv(
                data,
                dtype={"userId": np.int32, "movieId": np.int32, "tag": "string"},
            ).rename(columns={"userId": "user_id", "movieId": "item_id"})
            df["timestamp"] = pd.to_datetime(df["timestamp"], unit="s")
            df = df[df["tag"].notnull()]
            return df

    def genome_df(self):
        if isinstance(self.source, Path):
            if not (self.source / "genome-tags.csv").exists():
                return None
        else:
            name = self.prefix + "genome-tags.csv"
            if name not in self.source.namelist():
                return None

        with self.open_file("genome-tags.csv") as data:
            tags = pd.read_csv(
                data,
                dtype={"tagId": np.int32, "tag": "string"},
            )
        with self.open_file("genome-scores.csv") as data:
            scores = pd.read_csv(
                data, dtype={"movieId": np.int32, "tagId": np.int32, "relevance": np.float32}
            ).rename(columns={"movieId": "item_id"})

        tag_names = tags.set_index("tagId")["tag"].to_dict()
        df = pd.pivot(scores, columns="tagId", index="item_id", values="relevance")
        df = df.rename(columns=tag_names)
        return df

    def ratings_df(self):
        self._logger.debug("reading modern ratings CSV")
        with self.open_file("ratings.csv") as data:
            df = pd.read_csv(
                data,
                dtype={
                    "userId": np.int32,
                    "movieId": np.int32,
                    "rating": np.float32,
                    "timestamp": np.int64,
                },
            ).rename(columns={"userId": "user_id", "movieId": "item_id"})
            df["timestamp"] = pd.to_datetime(df["timestamp"], unit="s")
            return df


def load_movielens(path: str | Path) -> Dataset:
    """
    Load a MovieLens dataset.  The appropriate MovieLens format is detected
    based on the file contents.

    Stability:
        Caller

    Args:
        path:
            The path to the dataset, either as an unpacked directory or a zip
            file.

    Returns:
        The dataset.
    """
    with _ml_detect_and_open(path) as ml:
        return ml.dataset()


def load_movielens_df(path: str | Path) -> pd.DataFrame:
    """
    Load the ratings from a MovieLens dataset as a raw data frame.  The
    appropriate MovieLens format is detected based on the file contents.

    Stability:
        Caller

    Args:
        path:
            The path to the dataset, either as an unpacked directory or a zip
            file.

    Returns:
        The ratings, with columns ``user_id``, ``item_id``, ``rating``, and
        ``timestamp``.
    """
    with _ml_detect_and_open(path) as ml:
        return ml.ratings_df()


def _ml_detect_and_open(path: str | Path) -> MLData:
    loc = Path(path)
    version: str
    ctor: Callable[..., MLData]

    if loc.is_file() and loc.suffix == ".zip":
        log = _log.bind(zipfile=str(loc))
        log.debug("opening zip file")
        zf = ZipFile(loc, "r")
        try:
            infos = zf.infolist()
            first = infos[0]
            if not first.is_dir:  # pragma: nocover
                log.error("first entry is not directory")
                raise RuntimeError("invalid ML zip file")

            log.debug("base dir filename %s", first.filename)
            dsm = re.match(r"^(ml-(?:\d+[MmKk]|latest|latest-small))", first.filename)
            if not dsm:  # pragma: nocover
                log.error("invalid directory name %s", first.filename)
                raise RuntimeError("invalid ML zip file")

            version = dsm.group(1).lower()
            log.debug("found ML data set %s", version)
            ctor = MLData.version_impl(version)
            return ctor(version, zf, first.filename)
        except Exception as e:  # pragma nocover
            zf.close()
            raise e
    elif loc.is_dir():
        log = _log.bind(dir=str(loc))
        log.debug("loading from directory")
        dsm = re.match(r"^(ml-(?:\d+[MmKk]|latest(?:-small)?))", loc.name)
        if dsm:
            version = dsm.group(1)
            ctor = MLData.version_impl(dsm.group(1).lower())
            log.debug("inferred data set %s from dir name", version)
        else:
            log.debug("checking contents for data type")
            if (loc / "u.data").exists():
                log.debug("found u.data, interpreting as 100K")
                ctor = ML100KLoader
            elif (loc / "ratings.dat").exists():
                if (loc / "tags.dat").exists():
                    log.debug("found ratings.dat and tags.dat, interpreting as 10M")
                    version = "ml-10m"
                else:
                    log.debug("found ratings.dat but no tags, interpreting as 1M")
                    version = "ml-1m"
                ctor = MLMLoader
            elif (loc / "ratings.csv").exists():
                log.debug("found ratings.csv, interpreting as modern (20M and later)")
                version = "ml-modern"
                ctor = MLModernLoader
            else:
                log.error("could not detect MovieLens data")
                raise RuntimeError("invalid ML directory")

        return ctor(version, loc)

    elif loc.exists():  # pragma: nocover
        _log.error("invalid MovieLens data location", path=path)
        raise RuntimeError("not a directory or zip file")
    else:  # pragma: nocover
        _log.error("MovieLens data not found", path=path)
        raise FileNotFoundError(path)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from typing import Any

import numpy as np
import ray.tune

from lenskit.logging import get_logger

_log = get_logger(__name__)


class RelativePlateauStopper(ray.tune.Stopper):
    metric: str
    mode: str
    min_improvement: float
    check_iters: int
    grace_period: int

    results: dict[str, list[float]]

    def __init__(
        self,
        metric: str,
        mode: str,
        min_improvement: float = 0.01,
        check_iters: int = 3,
        grace_period: int = 5,
    ):
        assert check_iters <= grace_period, "check iters cannot be more than grace period"
        self.metric = metric
        self.mode = mode
        self.min_improvement = min_improvement
        self.check_iters = check_iters
        self.grace_period = grace_period
        self.results = {}

    def __call__(self, trial_id: str, result: dict[str, Any]) -> bool:
        epoch = result["training_iteration"]
        mr = result[self.metric]
        log = _log.bind(trial=trial_id, epoch=epoch, **{self.metric: mr})

        hist = self.results.get(trial_id, [])
        if len(hist) >= result["training_iteration"]:
            log.debug("truncating history", len=len(hist))
            hist = hist[: result["training_iteration"] - 1]
        hist.append(mr)
        self.results[trial_id] = hist
        if len(hist) < self.grace_period:
            log.debug("within grace period, accepting")
            return False

        hist = np.array(hist)
        # measure relative improvement over previous best entry
        if self.mode == "min":
            best = np.minimum.accumulate(hist)
            imp = hist[1:] / best[:-1] - 1.0
            imp *= -1
        else:
            best = np.maximum.accumulate(hist)
            imp = hist[1:] / best[:-1] - 1.0

        # if we haven't improved enough lately
        if np.all(imp[-self.check_iters :] < self.min_improvement).item():
            log.info(
                "trial plateaued, stopping with last improvement {:.3%}".format(imp[-1]),
                history=hist,
            )
            return True
        else:
            log.debug("continuing with improvement {:.3%}".format(imp[-1]))
            return False

    def stop_all(self) -> bool:
        return False



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Hyperparameter search.
"""

from __future__ import annotations

import os
from pathlib import Path
from typing import Literal

import click
import ray.tune.utils.log
from humanize import metric as human_metric
from humanize import precisedelta
from pydantic_core import to_json

from lenskit.logging import Task, get_logger, stdout_console
from lenskit.parallel import get_parallel_config
from lenskit.parallel.ray import init_cluster
from lenskit.tuning import PipelineTuner, TuningSpec

_log = get_logger(__name__)


@click.command("tune")
@click.option(
    "-T", "--training-data", type=Path, required=True, help="path to the training dataset"
)
@click.option(
    "-V",
    "--valid-data",
    "--tuning-data",
    "tuning_data",
    type=Path,
    required=True,
    help="path to the tuning/validation data",
)
@click.option("--random", "method", flag_value="random", help="use random search")
@click.option("--hyperopt", "method", flag_value="hyperopt", help="use HyperOpt search")
@click.option("--optuna", "method", flag_value="optuna", help="use Optuna search")
@click.option(
    "-j", "--job-limit", type=int, envvar="LK_TUNE_JOBS", help="limit for concurrent tuning jobs"
)
@click.option(
    "--max-points", type=int, metavar="N", help="maximum number of configurations to test"
)
@click.option(
    "-m",
    "--metric",
    type=click.Choice(["RMSE", "RBP", "RecipRank", "NDCG"]),
    default="RBP",
    help="the metric to optimize",
)
@click.option(
    "--save-pipeline", type=Path, metavar="FILE", help="Save best pipeline configuration to FILE"
)
@click.argument("SEARCH_SPEC", type=Path)
@click.argument("OUT", type=Path)
def tune(
    search_spec: Path,
    out: Path,
    method: Literal["random", "hyperopt", "optuna"] | None,
    max_points: int | None,
    job_limit: int | None,
    metric: str,
    training_data: Path,
    tuning_data: Path,
    save_pipeline: Path | None,
):
    """
    Tune pipeline hyperparameters with Ray Tune.
    """
    console = stdout_console()
    os.environ["RAY_AIR_NEW_OUTPUT"] = "0"
    ray.tune.utils.log.set_verbosity(0)

    spec = TuningSpec.load(search_spec)
    # override settings from command line
    if max_points is not None:
        spec.search.max_points = max_points
    if metric is not None:
        spec.search.metric = metric
    if method is not None:
        spec.search.method = method

    out.mkdir(exist_ok=True, parents=True)

    try:
        ray.init(address="auto")
        _log.info("connected to existing Ray cluster")
    except ConnectionError:
        # use global parallel setup for tuning
        init_cluster(global_logging=True, worker_parallel=get_parallel_config())

    # set up the tuning controller
    controller = PipelineTuner(spec, out)
    if job_limit is not None:
        controller.settings.jobs = job_limit
    controller.set_data(training_data, tuning_data)

    with Task(label=f"${spec.search.method} tune {controller.pipe_name}", tags=["tune"]) as task:
        task.save_to_file(out / "tuning-task.json")
        controller.setup()

        with open(out / "config.json", "wt") as jsf:
            print(controller.spec.model_dump_json(indent=2), file=jsf)

        _log.info("starting hyperparameter search")
        controller.run()

    best = controller.best_result()
    result_file = out / "result.json"
    result_json = to_json(best, indent=2)
    result_file.write_bytes(result_json + b"\n")

    if save_pipeline is not None:
        pipe_json = controller.best_pipeline().model_dump_json(indent=2)
        save_pipeline.write_text(pipe_json + "\n")

    _log.info("finished hyperparameter search")
    console.print("[bold yellow]Hyperparameter search completed![/bold yellow]")
    console.print("Best {} is [bold red]{:.3f}[/bold red]".format(metric, best[metric]))
    assert task.duration is not None
    line = "[bold magenta]{}[/bold magenta] trials took [bold cyan]{}[/bold cyan]".format(
        len(controller.results),
        precisedelta(task.duration),  # type: ignore
    )
    if task.system_power:
        line += " and consumed [bold green]{}[/bold green]".format(
            human_metric(task.system_power / 3600, unit="Wh")
        )
    console.print(line)
    console.print("Trial result:", best)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import sys
from io import StringIO
from pathlib import Path

import click
from rich.console import Console
from rich.markdown import Markdown

from lenskit.data.amazon import load_amazon_ratings
from lenskit.data.dataset import Dataset
from lenskit.data.movielens import load_movielens
from lenskit.data.summary import save_stats
from lenskit.logging import get_logger

_log = get_logger(__name__)


@click.command("describe")
@click.option("--movielens", "format", flag_value="movielens", help="describe MovieLens data")
@click.option("--amazon", "format", flag_value="amazon", help="describe Amazon rating data")
@click.option("--markdown", is_flag=True, help="output raw Markdown")
@click.argument("path", type=Path)
def describe(format: str | None, markdown: bool, path: Path):
    """
    Describe a data set.
    """

    log = _log.bind(path=str(path))

    match format:
        case None:
            log.info("loading LensKit native data")
            data = Dataset.load(path)
        case "movielens":
            log.info("loading MovieLens data")
            data = load_movielens(path)
        case "amazon":
            log.info("loading Amazon data")
            data = load_amazon_ratings(path)
        case _:
            raise ValueError(f"unknown data format {format}")

    console = Console()

    if markdown:
        save_stats(data, sys.stdout)
    else:
        out = StringIO()
        save_stats(data, out)
        console.print(Markdown(out.getvalue()), width=min(console.width, 80))



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import numpy as np
from numpy.typing import NDArray
from typing_extensions import Callable, TypeAlias, override

from lenskit.data import ItemList

from ._base import ListMetric, RankingMetricBase
from ._weighting import LogRankWeight, RankWeight

Discount: TypeAlias = Callable[[NDArray[np.number]], NDArray[np.float64]]


class DiscountWeight(RankWeight):
    def __init__(self, discount: Discount):
        self.discount = discount

    def weight(self, ranks):
        return np.reciprocal(np.maximum(self.discount(ranks), 1))


class NDCG(ListMetric, RankingMetricBase):
    """
    Compute the normalized discounted cumulative gain :cite:p:`ndcg`.

    Discounted cumultative gain is computed as:

    .. math::
        \\begin{align*}
        \\mathrm{DCG}(L,u) & = \\sum_{i=1}^{|L|} \\frac{r_{ui}}{d(i)}
        \\end{align*}

    Unrated items are assumed to have a utility of 0; if no rating values are
    provided in the truth frame, item ratings are assumed to be 1.

    This is then normalized as follows:

    .. math::
        \\begin{align*}
        \\mathrm{nDCG}(L, u) & = \\frac{\\mathrm{DCG}(L,u)}{\\mathrm{DCG}(L_{\\mathrm{ideal}}, u)}
        \\end{align*}

    Args:
        k:
            The maximum recommendation list length to consider (longer lists are
            truncated).
        weight:
            The rank weighting to use.
        discount:
            The discount function to use.  The default, base-2 logarithm, is the
            original function used by :cite:t:`ndcg`.  It is deprecated in favor
            of the ``weight`` option.
        gain:
            The field on the test data to use for gain values.  If ``None`` (the
            default), all items present in the test data have a gain of 1.  If
            set to a string, it is the name of a field (e.g. ``'rating'``).  In
            all cases, items not present in the truth data have a gain of 0.

    Stability:
        Caller
    """

    weight: RankWeight
    discount: Discount | None
    gain: str | None

    def __init__(
        self,
        k: int | None = None,
        *,
        weight: RankWeight = LogRankWeight(),
        discount: Discount | None = None,
        gain: str | None = None,
    ):
        super().__init__(k=k)
        self.weight = weight
        self.discount = discount
        if discount is not None:
            self.weight = DiscountWeight(discount)
        self.gain = gain

    @property
    def label(self):
        if self.k is not None:
            return f"NDCG@{self.k}"
        else:
            return "NDCG"

    @override
    def measure_list(self, recs: ItemList, test: ItemList) -> float:
        recs = self.truncate(recs)

        if self.gain:
            realized = _graded_dcg(recs, test, self.gain, self.weight)

            gains = test.field(self.gain, "pandas", index="ids")
            if gains is None:
                raise KeyError(f"test items have no field {self.gain}")
            if self.k:
                gains = gains.nlargest(n=self.k)
            else:
                gains = gains.sort_values(ascending=False)
            iweight = self.weight.weight(np.arange(1, len(gains) + 1))
            ideal = np.dot(gains.values, iweight).item()  # type: ignore

        else:
            realized = _binary_dcg(recs, test, self.weight)
            n = len(test)
            if self.k and self.k < n:
                n = self.k
            ideal = fixed_dcg(n, self.weight)

        return realized / ideal


class DCG(ListMetric, RankingMetricBase):
    """
    Compute the _unnormalized_ discounted cumulative gain :cite:p:`ndcg`.

    Discounted cumultative gain is computed as:

    .. math::
        \\begin{align*}
        \\mathrm{DCG}(L,u) & = \\sum_{i=1}^{|L|} \\frac{r_{ui}}{d(i)}
        \\end{align*}

    Unrated items are assumed to have a utility of 0; if no rating values are
    provided in the truth frame, item ratings are assumed to be 1.

    This metric does *not* normalize by ideal DCG. For that, use :class:`NDCG`.
    See :cite:t:`jeunenNormalisedDiscountedCumulative2024` for an argument for
    using the unnormalized version.

    Args:
        k:
            The maximum recommendation list length to consider (longer lists are
            truncated).
        discount:
            The discount function to use.  The default, base-2 logarithm, is the
            original function used by :cite:t:`ndcg`.
        gain:
            The field on the test data to use for gain values.  If ``None`` (the
            default), all items present in the test data have a gain of 1.  If set
            to a string, it is the name of a field (e.g. ``'rating'``).  In all
            cases, items not present in the truth data have a gain of 0.

    Stability:
        Caller
    """

    weight: RankWeight
    discount: Discount | None
    gain: str | None

    def __init__(
        self,
        k: int | None = None,
        *,
        weight: RankWeight = LogRankWeight(),
        discount: Discount | None = None,
        gain: str | None = None,
    ):
        super().__init__(k=k)
        self.weight = weight
        self.discount = discount
        if discount is not None:
            self.weight = DiscountWeight(discount)
        self.gain = gain

    @property
    def label(self):
        if self.k is not None:
            return f"DCG@{self.k}"
        else:
            return "DCG"

    @override
    def measure_list(self, recs: ItemList, test: ItemList) -> float:
        recs = self.truncate(recs)

        if self.gain:
            return _graded_dcg(recs, test, self.gain, self.weight)
        else:
            return _binary_dcg(recs, test, self.weight)


def _graded_dcg(
    recs: ItemList, test: ItemList, field: str, weight: RankWeight = LogRankWeight()
) -> float:
    gains = test.field(field, "pandas", index="ids")
    if gains is None:
        raise KeyError(f"test items have no field {field}")

    ranks = recs.ranks(format="pandas")
    if ranks is None:
        raise TypeError("item list is not ordered")

    ranks, gains = ranks.align(gains, join="inner")
    weights = weight.weight(ranks.values)  # type: ignore

    return np.dot(gains.values, weights)  # type: ignore


def _binary_dcg(recs: ItemList, test: ItemList, weight: RankWeight = LogRankWeight()) -> float:
    good = recs.isin(test)
    ranks = recs.ranks()
    if ranks is None:
        raise TypeError("item list is not ordered")

    weights = weight.weight(ranks[good])
    return np.sum(weights).item()


def array_dcg(
    scores: NDArray[np.number], weight: RankWeight = LogRankWeight(), *, graded: bool = True
):
    """
    Compute the Discounted Cumulative Gain of a series of recommended items with rating scores.
    These should be relevance scores; they can be :math:`{0,1}` for binary relevance data.

    This is not a true top-N metric, but is a utility function for other metrics.

    Args:
        scores:
            The utility scores of a list of recommendations, in recommendation order.
        discount:
            the rank discount function.  Each item's score will be divided the discount of its rank,
            if the discount is greater than 1.

    Returns:
        double: the DCG of the scored items.
    """
    ids = np.arange(1, len(scores) + 1)
    recs = ItemList(item_ids=ids, ordered=True)

    mask = scores > 0
    test = ItemList(item_ids=ids[mask], rating=scores[mask])

    if graded:
        return _graded_dcg(recs, test, "rating")
    else:
        return _binary_dcg(recs, test)


def fixed_dcg(n: int, weight: RankWeight = LogRankWeight()) -> float:
    """
    Compute the Discounted Cumulative Gain of a fixed number of items with
    relevance 1.
    """

    ranks = np.arange(1, n + 1)
    wvec = weight.weight(ranks)
    return np.sum(wvec).item()



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from pathlib import Path

from lenskit.als import BiasedMFScorer, ImplicitMFScorer
from lenskit.basic import BiasScorer, FallbackScorer, TopNRanker
from lenskit.config import load_config_data
from lenskit.pipeline import Pipeline, PipelineConfig
from lenskit.pipeline.nodes import ComponentInstanceNode

config_dir = Path("pipelines")


def test_load_config():
    als_file = config_dir / "als-implicit.toml"
    als_cfg = load_config_data(als_file, PipelineConfig)
    assert isinstance(als_cfg, PipelineConfig)
    assert als_cfg.meta.name == "als-implicit"


def test_apply_scorer_config():
    als_file = config_dir / "als-implicit.toml"
    als_pipe = Pipeline.load_config(als_file)

    node = als_pipe.node("scorer")
    assert isinstance(node, ComponentInstanceNode)
    assert isinstance(node.component, ImplicitMFScorer)

    node = als_pipe.node("recommender")
    assert isinstance(node, ComponentInstanceNode)
    assert isinstance(node.component, TopNRanker)


def test_apply_predictor_config():
    als_file = config_dir / "als-explicit.toml"
    als_pipe = Pipeline.load_config(als_file)

    node = als_pipe.node("scorer")
    assert isinstance(node, ComponentInstanceNode)
    assert isinstance(node.component, BiasedMFScorer)

    node = als_pipe.node("recommender")
    assert isinstance(node, ComponentInstanceNode)
    assert isinstance(node.component, TopNRanker)

    node = als_pipe.node("rating-predictor")
    assert isinstance(node, ComponentInstanceNode)
    assert isinstance(node.component, FallbackScorer)

    node = als_pipe.node("fallback-predictor")
    assert isinstance(node, ComponentInstanceNode)
    assert isinstance(node.component, BiasScorer)


def test_apply_no_fallback():
    config = {
        "meta": {"name": "bias"},
        "options": {"base": "std:topn-predict", "fallback_predictor": False},
        "components": {"scorer": {"class": "lenskit.basic.BiasScorer"}},
    }
    bias_pipe = Pipeline.from_config(config)

    node = bias_pipe.node("scorer")
    assert isinstance(node, ComponentInstanceNode)
    assert isinstance(node.component, BiasScorer)

    node = bias_pipe.node("recommender")
    assert isinstance(node, ComponentInstanceNode)
    assert isinstance(node.component, TopNRanker)

    node = bias_pipe.node("rating-predictor")
    assert isinstance(node, ComponentInstanceNode)
    assert isinstance(node.component, BiasScorer)

    node = bias_pipe.node("fallback-predictor", missing=None)
    assert node is None


def test_apply_config():
    config = {
        "meta": {"name": "als-large"},
        "options": {"base": "std:topn", "default_length": 100},
        "components": {
            "scorer": {"class": "lenskit.als.ImplicitMFScorer", "config": {"embedding_size": 256}}
        },
    }
    als_pipe = Pipeline.from_config(config)

    node = als_pipe.node("scorer")
    assert isinstance(node, ComponentInstanceNode)
    assert isinstance(node.component, ImplicitMFScorer)
    assert node.component.config.embedding_size == 256

    node = als_pipe.node("recommender")
    assert isinstance(node, ComponentInstanceNode)
    assert isinstance(node.component, TopNRanker)
    assert node.component.config.n == 100



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import numpy as np
import pyarrow as pa
import torch

import hypothesis.extra.numpy as nph
import hypothesis.strategies as st
from hypothesis import given

from lenskit.data.arrow import arrow_to_format


@given(nph.arrays(dtype=nph.floating_dtypes(endianness="="), shape=nph.array_shapes(max_dims=1)))
def test_array_to_arrow(data: np.ndarray):
    arr = pa.array(data)

    arr_arrow = arrow_to_format(arr, "arrow")
    assert arr_arrow is arr


@given(nph.arrays(dtype=nph.floating_dtypes(endianness="="), shape=nph.array_shapes(max_dims=1)))
def test_array_to_numpy(data: np.ndarray):
    arr = pa.array(data)

    arr_numpy = arrow_to_format(arr, "numpy")
    assert isinstance(arr_numpy, np.ndarray)
    assert len(arr_numpy) == len(data)


@given(nph.arrays(dtype=nph.floating_dtypes(endianness="="), shape=nph.array_shapes(max_dims=1)))
def test_array_to_torch(data: np.ndarray):
    arr = pa.array(data)

    arr_torch = arrow_to_format(arr, "torch")
    assert isinstance(arr_torch, torch.Tensor)
    assert len(arr_torch) == len(data)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Recommender systems toolkit.
"""

import lazy_loader as lazy

from ._version import lenskit_version

__version__ = lenskit_version()


# lazy-load LensKit internal imports (per SPEC-1)
# IMPORTANT: this must be kept in sync with __init__.pyi
__getattr__, __dir__, __all__ = lazy.attach(
    __name__,
    submodules=[
        # modules
        "batch",
        "config",
        "data",
        "logging",
        "metrics",
        "pipeline",
        "operations",
        "splitting",
        "stats",
        "torch",
        "training",
        # component modules
        "als",
        "basic",
        "flexmf",
        "funksvd",
        "hpf",
        "knn",
        "sklearn",
    ],
    submod_attrs={
        "config": ["configure", "lenskit_config"],
        "data": ["Dataset", "DatasetBuilder"],
        "operations": ["predict", "recommend", "score"],
        "pipeline": ["Pipeline", "RecPipelineBuilder", "Component", "topn_pipeline"],
        "splitting": ["TTSplit"],
    },
)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from __future__ import annotations

from typing import Literal

import numpy as np
import torch
from pydantic import BaseModel, PositiveInt

from lenskit.config.common import EmbeddingSizeMixin
from lenskit.data import ItemList, QueryInput, RecQuery, Vocabulary
from lenskit.logging import get_logger
from lenskit.pipeline import Component
from lenskit.torch import inference_mode
from lenskit.training import UsesTrainer

from ._model import FlexMFModel

# I want a logger for information
_log = get_logger(__name__)


class FlexMFConfigBase(EmbeddingSizeMixin, BaseModel):
    """
    Common configuration for all FlexMF scoring components.

    Stability:
        Experimental
    """

    embedding_size: PositiveInt = 64

    batch_size: int = 8 * 1024
    """
    The training batch size.
    """

    learning_rate: float = 0.01
    """
    The learning rate for training.
    """

    epochs: int = 10
    """
    The number of training epochs.
    """

    regularization: float = 0.01
    """
    The regularization strength.

    .. note::
        The explicit-feedback model uses a different default strength.
    """

    reg_method: Literal["AdamW", "L2"] | None = "AdamW"
    """
    The regularization method to use.

    With the default AdamW regularization, training will use the
    :class:`~torch.optim.AdamW` optimizer with weight decay. With L2
    regularization, training will use sparse gradients and the
    :class:`torch.optim.SparseAdam` optimizer.

    .. note::
        The explicit-feedback model defaults this setting to ``"L2"``.

    ``None``
        Use no regularization.

    ``"L2"``
        Use L2 regularization on the parameters used in each training batch. The
        strength is applied to the _mean_ norms in a batch, so that the
        regularization term scale is not dependent on the batch size.

    ``"AdamW"``
        Use :class:`torch.optim.AdamW` with the specified regularization
        strength.  This configuration does *not* use sparse gradients, but
        training time is often comparable.

    .. note::
        Regularization values do not necessarily have the same range or meaning
        for the different regularization methods.
    """


class FlexMFScorerBase(UsesTrainer, Component):
    """
    Base class for the FlexMF scorers, providing common Torch support.

    Stability:
        Experimental
    """

    config: FlexMFConfigBase
    users: Vocabulary
    items: Vocabulary
    model: FlexMFModel

    def to(self, device):
        "Move the model to a different device."
        self.model = self.model.to(device)
        return self

    @inference_mode
    def __call__(self, query: QueryInput, items: ItemList) -> ItemList:
        """
        Generate item scores for a user.

        Note that user and items are both user and item IDs, not positions.
        """
        # make sure the query is in a known / usable format
        query = RecQuery.create(query)

        # if we have no user ID, we cannot score items
        # TODO: support pooling from user history
        u_row = None
        if query.user_id is not None:
            u_row = self.users.number(query.user_id, missing=None)

        if u_row is None:
            return ItemList(items, scores=np.nan)

        # look up the user row in the embedding matrix
        u_tensor = torch.IntTensor([u_row])
        # make sure it's on the right device
        device = self.model.device
        u_tensor = u_tensor.to(device, non_blocking=True)

        # look up the item columns in the embedding matrix
        i_cols = items.numbers(vocabulary=self.items, missing="negative", format="torch")
        i_cols = i_cols.to(device, non_blocking=True)

        # unknown items will have column -1 - limit to the
        # ones we know, and remember which item IDs those are
        scorable_mask = i_cols.ge(0)
        i_cols = i_cols.masked_select(scorable_mask)

        # get scores
        scores = self.score_items(u_tensor, i_cols)
        # initialize output score array, fill with missing
        full_scores = torch.full((len(items),), np.nan, dtype=torch.float32, device=scores.device)
        full_scores.masked_scatter_(scorable_mask, scores)

        # return the result!
        return ItemList(items, scores=full_scores)

    def score_items(self, users: torch.Tensor, items: torch.Tensor) -> torch.Tensor:
        """
        Score for users and items, after resolivng them and limiting to known
        users and items.
        """
        return self.model(users, items)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Support for arbitrary record sinks.
"""

from typing import Generic, Protocol, TypeVar
from uuid import UUID

R = TypeVar("R", contravariant=True)


class RecordSink(Protocol, Generic[R]):
    """
    Generic interface for record sinks.

    Stability:
        Internal
    """

    sink_id: UUID

    def record(self, record: R, /):
        """
        Record the timings for a pipeline run.
        """



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from uuid import uuid4

import numpy as np
import pandas as pd
import pyarrow as pa
import pyarrow.compute as pc

from pytest import importorskip, mark

from lenskit.data import ItemList, Vocabulary

numexpr = importorskip("numexpr")

LEFT_SIZE = 50000
RIGHT_SIZE = 100
TOTAL_SIZE = LEFT_SIZE * 2


@mark.benchmark(group="geometric")
@mark.parametrize("size", [100, 1000, 50000])
def test_numpy_geom(benchmark, size: int):
    ranks = np.arange(1, size + 1)

    def search():
        _w = np.power(0.85, ranks - 1)

    benchmark(search)


@mark.benchmark(group="geometric")
@mark.parametrize("size", [100, 1000, 50000])
def test_numpy_geom_log(benchmark, size: int):
    ranks = np.arange(1, size + 1)

    def search():
        _w = np.exp(np.log(0.85) * (ranks - 1))

    benchmark(search)


@mark.benchmark(group="geometric")
@mark.parametrize("size", [100, 1000, 50000])
def test_numpy_geom_log_ip(benchmark, size: int):
    ranks = np.arange(1, size + 1)

    def search():
        scores = np.require(ranks, dtype=np.float64)
        scores -= 1
        scores *= np.log(0.85)
        np.exp(scores, out=scores)

    benchmark(search)


@mark.benchmark(group="geometric")
@mark.parametrize("size", [100, 1000, 50000])
def test_numexpr_geom(benchmark, size: int):
    ranks = np.arange(1, size + 1)

    def search():
        _w = numexpr.evaluate("patience ** (rank - 1)", {"patience": 0.85, "rank": ranks})

    benchmark(search)


@mark.benchmark(group="geometric")
@mark.parametrize("size", [100, 1000, 50000])
def test_numexpr_geom_log(benchmark, size: int):
    ranks = np.arange(1, size + 1)

    def search():
        _w = numexpr.evaluate("exp(log(patience) * (rank - 1))", {"patience": 0.85, "rank": ranks})

    benchmark(search)


@mark.benchmark(group="logarithmic")
@mark.parametrize("size", [100, 1000, 50000])
def test_numpy_log_old(benchmark, size: int):
    ranks = np.arange(1, size + 1)

    def search():
        _w = np.reciprocal(np.log(np.maximum(ranks, 2)) / np.log(2))

    benchmark(search)


@mark.benchmark(group="logarithmic")
@mark.parametrize("size", [100, 1000, 50000])
def test_numpy_log(benchmark, size: int):
    ranks = np.arange(1, size + 1)

    def search():
        _w = np.log(2) / np.log(np.maximum(ranks, 2))

    benchmark(search)


@mark.benchmark(group="logarithmic")
@mark.parametrize("size", [100, 1000, 50000])
def test_numexpr_log(benchmark, size: int):
    ranks = np.arange(1, size + 1)

    def search():
        _w = numexpr.evaluate(
            "log(base) / log(where(rank < 2, 2, rank))", {"base": 2, "rank": ranks}
        )

    benchmark(search)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
LightGCN recommendation.
"""

from __future__ import annotations

import warnings

import numpy as np
import torch
from pydantic import BaseModel, PositiveFloat, PositiveInt, model_validator
from structlog.stdlib import BoundLogger
from torch_geometric.nn import LightGCN
from typing_extensions import Literal, Self

from lenskit import logging
from lenskit.config.common import EmbeddingSizeMixin
from lenskit.data import BatchIter, Dataset, ItemList, QueryInput, RecQuery, Vocabulary
from lenskit.data.matrix import COOStructure
from lenskit.data.relationships import MatrixRelationshipSet
from lenskit.logging.progress._dispatch import item_progress
from lenskit.pipeline.components import Component
from lenskit.torch import inference_mode
from lenskit.training import ModelTrainer, TrainingOptions, UsesTrainer

_log = logging.get_logger(__name__)


class LightGCNConfig(EmbeddingSizeMixin, BaseModel):
    """
    Configuration for :class:`LightGCNScorer`.

    Stability:
        Experimental
    """

    embedding_size: PositiveInt = 16
    """
    The dimension of the embedding space (number of latent features).  Seems
    to work best as a power of 2.
    """

    layer_count: PositiveInt = 2
    """
    The number of layers to use.
    """

    layer_blend: PositiveFloat | list[PositiveFloat] | None = None
    """
    The blending coefficient(s) for layer blending.  This is equivalent to
    ``alpha`` in :class:`LightGCN`.
    """

    batch_size: PositiveInt = 4 * 1024
    """
    The training batch size.
    """

    learning_rate: PositiveFloat = 0.01
    """
    The learning rate for training.
    """

    epochs: PositiveInt = 10
    """
    The number of training epochs.
    """

    regularization: PositiveFloat | None = 0.01
    """
    The regularization strength.
    """

    loss: Literal["logistic", "pairwise"] = "pairwise"
    """
    The loss to use for model training.

    ``pairwise``
        BPR pairwise ranking loss, using :meth:`LightGCN.recommend_loss`.

    ``logistic``
        Logistic link prediction loss, using :meth:`LightGCN.link_pred_loss`.
    """

    @model_validator(mode="after")
    def check_layer_blending(self) -> Self:
        if isinstance(self.layer_blend, list) and len(self.layer_blend) != self.layer_count:
            raise ValueError(
                f"layer_blend has length {len(self.layer_blend)}, expected {self.layer_count}"
            )

        return self


class LightGCNScorer(UsesTrainer, Component[ItemList]):
    """
    Scorer using :class:`LightGCN` :cite:p:`heLightGCNSimplifyingPowering2020`.

    Stability:
        Experimental
    """

    config: LightGCNConfig

    users: Vocabulary
    items: Vocabulary
    model: LightGCN
    _user_base: int
    """
    Offset for the user nodes in the data set. Item edges start from zero.
    """
    _edges: torch.Tensor
    """
    The graph edges as used by the graph model.
    """

    def to(self, device):
        "Move the model to a different device."
        self.model = self.model.to(device)
        self._edges = self._edges.to(device)
        return self

    @inference_mode
    def __call__(self, query: QueryInput, items: ItemList) -> ItemList:
        """
        Generate item scores for a user.

        Note that user and items are both user and item IDs, not positions.
        """
        # make sure the query is in a known / usable format
        query = RecQuery.create(query)

        # if we have no user ID, we cannot score items
        u_row = None
        if query.user_id is not None:
            u_row = self.users.number(query.user_id, missing=None)

        if u_row is None:
            return ItemList(items, scores=np.nan)

        # look up the item columns in the embedding matrix
        i_cols = items.numbers(vocabulary=self.items, missing="negative", format="torch")
        i_cols = i_cols.to(self._edges.device, dtype=self._edges.dtype)

        # unknown items will have column -1 - limit to the
        # ones we know, and remember which item IDs those are
        scorable_mask = i_cols.ge(0)
        i_cols = i_cols.masked_select(scorable_mask)

        # set up the edge tensor
        u_tensor = torch.from_numpy(np.repeat(np.array([u_row + self._user_base]), len(i_cols)))
        u_tensor = u_tensor.to(self._edges.device, dtype=self._edges.dtype)
        edges = torch.stack([u_tensor, i_cols])
        scores = self.model(self._edges, edges)

        # initialize output score array, fill with missing
        full_scores = torch.full((len(items),), np.nan, dtype=torch.float32, device=scores.device)
        full_scores.masked_scatter_(scorable_mask, scores)

        # return the result!
        return ItemList(items, scores=full_scores)

    def create_trainer(self, data, options):
        match self.config.loss:
            case "logistic":
                return LogisticLightGCNTrainer(self, data, options)
            case "pairwise":
                return PairwiseLightGCNTrainer(self, data, options)
            case _:  # pragam: nocover
                raise ValueError("invalid loss")


class LightGCNTrainer(ModelTrainer):
    scorer: LightGCNScorer
    data: Dataset
    options: TrainingOptions
    log: BoundLogger

    rng: np.random.Generator
    device: str
    model: LightGCN

    matrix: MatrixRelationshipSet
    coo: COOStructure
    user_base: int
    edges: torch.Tensor

    optimizer: torch.optim.Optimizer
    epochs_trained: int = 0

    def __init__(self, scorer: LightGCNScorer, data: Dataset, options: TrainingOptions):
        self.scorer = scorer
        self.data = data
        self.options = options
        self.device = options.configured_device(gpu_default=True)
        self.log = _log.bind()
        self.rng = options.random_generator()

        _log.info("initializing LightGCN trainer", device=str(self.device))

        self.user_base = data.item_count
        node_count = data.user_count + data.item_count
        if isinstance(scorer.config.layer_blend, list):
            blend = torch.tensor(scorer.config.layer_blend)
        else:
            blend = scorer.config.layer_blend

        self.matrix = data.interactions().matrix()
        self.coo = coo = self.matrix.coo_structure()
        e_src = torch.tensor(coo.row_numbers + self.user_base)
        e_dst = torch.tensor(coo.col_numbers)

        edge_dtype = torch.int64
        if torch.device(self.device).type == "cuda":
            if torch.__version__ >= "2.8":
                edge_dtype = torch.int32
            else:
                warnings.warn("using 64-bit edge indices, upgrade to Torch 2.8 for 32-bit")

        self.edges = torch.stack([e_src, e_dst]).to(self.device, dtype=edge_dtype)

        self.model = LightGCN(
            node_count, scorer.config.embedding_size, scorer.config.layer_count, blend
        ).to(self.device)

        if scorer.config.regularization:
            self.optimizer = torch.optim.AdamW(
                self.model.parameters(),
                scorer.config.learning_rate,
                weight_decay=scorer.config.regularization,
            )
        else:
            self.optimizer = torch.optim.Adam(
                self.model.parameters(),
                scorer.config.learning_rate,
            )

        self.scorer.items = self.data.items
        self.scorer.users = self.data.users
        self.scorer.model = self.model
        self.scorer._user_base = self.user_base
        self.scorer._edges = self.edges

    def train_epoch(self) -> dict[str, float] | None:
        epoch = self.epochs_trained + 1
        self.log = elog = self.log.bind(epoch=epoch)
        elog.debug("creating epoch training data")
        config = self.scorer.config

        # permute the data
        perm = np.require(self.rng.permutation(self.coo.nnz), dtype=np.int32)
        perm_t = torch.from_numpy(perm).to(self.device)

        batches = BatchIter(self.coo.nnz, config.batch_size)

        tot_loss = torch.tensor(0.0).to(self.device)
        avg_loss = np.nan
        with item_progress(f"Training epoch {epoch}", len(batches), {"loss": ".3f"}) as pb:
            elog.debug("beginning epoch")
            for i, (bs, be) in enumerate(batches, start=1):
                self.log = blog = elog.bind(batch=i)
                blog.debug("training batch")

                pos = self.edges[:, perm_t[bs:be]]
                neg = self.matrix.sample_negatives(
                    self.coo.row_numbers[perm[bs:be]],
                    rng=self.rng,
                )
                neg = torch.from_numpy(neg)
                neg = neg.to(self.device, dtype=pos.dtype, non_blocking=True)
                neg = torch.stack([pos[0], neg])

                mb_edges = torch.cat([pos, neg], 1)
                assert mb_edges.shape == (2, (be - bs) * 2)

                scores = self.model(self.edges, mb_edges)
                loss = self.batch_loss(mb_edges, scores)

                loss.backward()
                self.optimizer.step()

                if i % 20 == 0:
                    avg_loss = tot_loss.item() / i
                pb.update(loss=avg_loss)
                tot_loss += loss

        avg_loss = tot_loss.item() / len(batches)
        elog.debug("epoch complete", loss=avg_loss)
        self.epochs_trained += 1
        return {"loss": avg_loss}

    def finalize(self):
        pass

    def batch_loss(self, mb_edges: torch.Tensor, scores: torch.Tensor) -> torch.Tensor:
        raise NotImplementedError()


class LogisticLightGCNTrainer(LightGCNTrainer):
    def batch_loss(self, mb_edges: torch.Tensor, scores: torch.Tensor):
        (n,) = scores.shape
        labels = torch.from_numpy(np.repeat([1, 0], n // 2)).to(self.device)
        return self.model.link_pred_loss(scores, labels)


class PairwiseLightGCNTrainer(LightGCNTrainer):
    def batch_loss(self, mb_edges: torch.Tensor, scores: torch.Tensor):
        (n,) = scores.shape
        pos_score, neg_score = scores.chunk(2)
        # FIXME: set up better regularization
        return self.model.recommendation_loss(pos_score, neg_score, node_id=mb_edges.ravel())



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from lenskit.basic import BiasScorer, PopScorer
from lenskit.pipeline.common import topn_pipeline


def test_merge_simple_setting():
    pipe = topn_pipeline(PopScorer, n=100)
    cfg = pipe.config
    assert cfg.components["scorer"].config["score"] == "quantile"

    c2 = pipe.config.merge_component_configs({"scorer": {"score": "rank"}})
    assert c2 is not cfg
    assert c2.components["scorer"].config["score"] == "rank"


def test_merge_deep_setting():
    pipe = topn_pipeline(BiasScorer, n=100)
    cfg = pipe.config
    assert cfg.components["scorer"].config["damping"] == 0.0

    c2 = pipe.config.merge_component_configs({"scorer": {"damping": {"user": 150, "item": 4}}})
    assert c2 is not cfg
    assert c2.components["scorer"].config["damping"]["user"] == 150
    assert c2.components["scorer"].config["damping"]["item"] == 4



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
FA*IR top-N re-ranking.
"""

# pyright: basic
from __future__ import annotations

from collections import deque

import numpy as np
from pydantic import BaseModel, Field, PositiveFloat, PositiveInt
from scipy.stats import binom
from typing_extensions import override

from lenskit.data import Dataset, ItemList
from lenskit.logging import Stopwatch, get_logger
from lenskit.pipeline import Component
from lenskit.training import Trainable, TrainingOptions

_log = get_logger(__name__)


class FAIRRerankerConfig(BaseModel, extra="forbid"):
    """
    Configuration for :class:`FAIRReranker`.
    """

    n: PositiveInt
    """
    Size of the top-N subset to rerank. If ``None``, the entire
    candidate list is reranked.
    """

    p: PositiveFloat = Field(0.5, gt=0.0, lt=1.0)
    """
    The target proportion of protected items in the list (0.0 ≤ p ≤ 1.0).
    The default value is 0.5.
    """

    alpha: PositiveFloat = Field(0.1, gt=0.0, lt=1.0)
    """
    The probability threshold for Type I error across prefixes.
    The default value is 0.1.
    """

    protected_attribute: str = Field(
        "protected", description="Item attribute name for the protected flag"
    )
    """
    The item attribute name that marks whether an item is protected.
    The default name is 'protected'.
    """


class FAIRReranker(Component[ItemList], Trainable):
    """
    FA*IR top-N re-ranking algorithm.

    Re-ranks a sorted candidate list to satisfy binomial lower-bound constraints
    for a binary protected/unprotected attribute at every prefix (1..N). It expects
    a sorted list of items, the target proportion of protected items ``p`` and
    significance level ``alpha``.

    For details, see `Zehlike et al. (2017)`_.

    .. _Zehlike et al. (2017): https://doi.org/10.1145/3132847.3132938

    Stability:
        Caller

    """

    config: FAIRRerankerConfig
    "FA*IR reranker configuration."

    def _compute_m_list(self, n, p, alpha):
        n_vals = np.arange(1, n + 1)
        # if we see m or fewer protected, it’s rare enough (probability ≤ α) that we call it unfair.
        m_list_ = binom.ppf(alpha, n_vals, p)
        # set bounds for edge cases (<0 or >N)
        m_list_ = np.clip(m_list_, 0, n_vals).astype(int)
        return m_list_

    def _compute_blocks(self, m_list_):
        max_m = int(m_list_[-1])
        if max_m == 0:
            return []

        # find the positions where m increases to form blocks
        # (last value of each block increases by 1)
        change_points = np.flatnonzero(np.diff(m_list_, prepend=0)) + 1
        block_sizes = np.diff(change_points, prepend=0)
        return block_sizes

    def _compute_rejection_prob(self, n, p, alpha_c):
        """
        Algorithm 1 (AdjustSignificance), exactly:
        - compute m(pos) from Binomial PPF at alpha_c
        - compute blocks b(j)
        - for j=1..m(n):
            S = conv(S, Bin(b(j), p))
            S[j-1] = 0     # prune bucket j-1
        Returns rejection probability = 1- sum(S).
        """
        m_list_ = self._compute_m_list(n, p, alpha_c)
        m_blocks = self._compute_blocks(m_list_)
        S = np.array([1.0], dtype=np.float64)

        if len(m_blocks) > 0:
            for j, block_size in enumerate(m_blocks, start=1):
                if block_size not in self.pmf_cache:
                    self.pmf_cache[block_size] = binom.pmf(np.arange(block_size + 1), block_size, p)

                S = np.convolve(self.pmf_cache[block_size], S)
                # set the next index (j-1) to zero since the last element of each block
                # increases by 1 and now we need at least j protected up to this point
                S[j - 1] = 0

        return float(1 - S.sum())

    def _binary_search_significance(self, n, p, alpha, tolerance=1e-10, max_iter=100):
        low, high = 0, 1

        for _ in range(max_iter):
            alpha_c = (low + high) / 2
            rejection_prob = self._compute_rejection_prob(n, p, alpha_c)

            if rejection_prob > alpha:
                high = alpha_c
            else:
                low = alpha_c
            if abs(rejection_prob - alpha) < tolerance or high - low < tolerance:
                break

        alpha_c = (low + high) / 2
        return alpha_c

    @override
    def train(self, data: Dataset, options: TrainingOptions = TrainingOptions()):
        """
        Precompute adjusted alpha and m-table for the configured (n,p,alpha)

        Training computes the (multiple-tests–corrected) prefix requirements m[i]
        so that every prefix of size i has at least m[i] protected items with
        significance α (after correction).

        Args:
            protected attributes:
                The item entity table must have a scalar attribute ``protected`` that is ``True``
                if the item is in the protected group and ``False`` otherwise.
        """
        self.pmf_cache = {}

        log = _log.bind(alpha=self.config.alpha, n_max=self.config.n)
        timer = Stopwatch()
        log.info("computing FA*IR thresholds")

        # 1. find α_c
        self.alpha_c_ = self._binary_search_significance(
            n=self.config.n, p=self.config.p, alpha=self.config.alpha
        )
        # 2. compute m_table
        self.m_list_ = self._compute_m_list(n=self.config.n, p=self.config.p, alpha=self.alpha_c_)

        items = data.entities("item")
        protected = self.config.protected_attribute

        if protected not in items.attributes:
            raise ValueError(
                f"Dataset is missing required '{protected}' attribute for item entities"
            )

        self.protected_attributes = items.attribute(protected).pandas()

        log.info(
            "[%s] computed thresholds with p=%.2f, alpha=%.2f → alpha_c=%.8f",
            timer,
            self.config.p,
            self.config.alpha,
            self.alpha_c_,
        )

    def __call__(self, items: ItemList, n: int | None = None) -> ItemList:
        item_ids = items.ids()

        is_protected = self.protected_attributes.reindex(item_ids).fillna(False).values

        p_items = deque(np.nonzero(is_protected)[0])
        up_items = deque(np.nonzero(~is_protected)[0])

        count_prot = 0
        reranked_indices = []
        list_size = len(item_ids)
        n_config = self.config.n

        # check that runtime n <= configured n
        if n is not None and n > n_config:
            raise ValueError(f"The requested rerank length n={n}, exceeds configured n={n_config}.")
        # use provided n or fallback to config.n (never larger than list size)
        n = min(n or n_config, list_size)

        # rerank the list
        for i in range(n):
            if count_prot < self.m_list_[i]:
                if p_items:
                    reranked_indices.append(p_items.popleft())
                    count_prot += 1
                else:
                    reranked_indices.extend(up_items)
                    break
            else:
                if up_items:
                    reranked_indices.append(up_items.popleft())
                else:
                    reranked_indices.extend(p_items)
                    count_prot += len(p_items)
                    break

        topn_reranked = ItemList(items[reranked_indices], ordered=True)
        return topn_reranked
